from skibidi_orm.exceptions.operations import IrreversibleOperationError
import skibidi_orm.migration_engine.operations.table_operations as t_ops
import skibidi_orm.migration_engine.operations.column_operations as c_ops
from skibidi_orm.migration_engine.operations.operation_type import OperationType
import skibidi_orm.migration_engine.adapters.database_objects.constraints as c
from skibidi_orm.migration_engine.adapters.base_adapter import BaseTable, BaseColumn
from typing import Literal
from pytest import raises

DataType = Literal["TEXT", "INTEGER", "REAL", "BLOB", "NULL"]
mock_table_1 = BaseTable[BaseColumn[DataType]]("table")
mock_table_2 = BaseTable[BaseColumn[DataType]]("table_2")
mock_column_1 = BaseColumn[DataType]("column", "INTEGER")
mock_column_2 = BaseColumn[DataType]("column_2", "TEXT")


def test_create_table_operation_init_reverse():
    operation = t_ops.CreateTableOperation(table=mock_table_1)
    assert operation.operation_type == OperationType.CREATE
    assert operation.table == mock_table_1
    assert operation.is_reversible is True

    reverse_operation = operation.reverse()
    assert reverse_operation == t_ops.DeleteTableOperation(table=operation.table)


def test_delete_table_operation_init_reverse():
    operation = t_ops.DeleteTableOperation(table=mock_table_1)
    assert operation.operation_type == OperationType.DELETE
    assert operation.table == mock_table_1
    assert operation.is_reversible is False

    with raises(IrreversibleOperationError):
        operation.reverse()


def test_rename_table_operation_init_reverse():
    operation = t_ops.RenameTableOperation(
        table=mock_table_1, new_name=mock_table_2.name
    )
    assert operation.operation_type == OperationType.RENAME
    assert operation.table == mock_table_1
    assert operation.new_name == mock_table_2.name
    assert operation.is_reversible is True
    assert operation.reverse() == t_ops.RenameTableOperation(
        table=mock_table_1, new_name=mock_table_1.name
    )


def test_add_column_operation_init_reverse():
    operation = c_ops.AddColumnOperation(table=mock_table_1, column=mock_column_1)
    assert operation.operation_type == OperationType.CREATE
    assert operation.table == mock_table_1
    assert operation.column == mock_column_1
    assert operation.is_reversible is True

    reverse_operation = operation.reverse()
    assert reverse_operation == c_ops.DeleteColumnOperation(
        table=operation.table,
        column=operation.column,
    )


def test_add_column_composite_foreign_key():
    """Shouldn't be able to initialise an c_ops.AddColumnOperation with a composite foreign key."""
    with raises(ValueError):
        c_ops.AddColumnOperation(
            table=mock_table_1,
            column=mock_column_1,
            related_foreign_key=c.ForeignKeyConstraint(
                mock_table_1.name,
                mock_table_2.name,
                {
                    "composite": "key",
                    mock_column_1.name: mock_column_2.name,
                },
            ),
        )


def test_add_column_foreign_key_references_different_column():
    """Shouldn't be able to initialise an c_ops.AddColumnOperation with a foreign key that references a different column."""
    with raises(ValueError):
        c_ops.AddColumnOperation(
            table=mock_table_1,
            column=mock_column_1,
            related_foreign_key=c.ForeignKeyConstraint(
                mock_table_1.name,
                mock_table_2.name,
                {
                    mock_column_2.name: "this should throw an exception",
                },
            ),
        )


def test_add_column_foreign_key_references_different_table():
    """Shouldn't be able to initialise an c_ops.AddColumnOperation with a foreign key that references the wrong table."""
    with raises(ValueError):
        c_ops.AddColumnOperation(
            table=mock_table_1,
            column=mock_column_1,
            related_foreign_key=c.ForeignKeyConstraint(
                mock_table_2.name,
                "this should throw an exception",
                {
                    mock_column_1.name: mock_column_2.name,
                },
            ),
        )


def test_delete_column_operation_init_reverse():
    operation = c_ops.DeleteColumnOperation(table=mock_table_1, column=mock_column_1)
    assert operation.operation_type == OperationType.DELETE
    assert operation.table == mock_table_1
    assert operation.column == mock_column_1
    assert operation.is_reversible is False
    with raises(IrreversibleOperationError):
        operation.reverse()


def test_rename_column_operation_init_reverse():
    operation = c_ops.RenameColumnOperation(
        table=mock_table_1,
        column=mock_column_1,
        new_name=mock_column_2.name,
    )
    assert operation.operation_type == OperationType.RENAME
    assert operation.table == mock_table_1
    assert operation.column == mock_column_1
    assert operation.new_name == mock_column_2.name
    assert operation.is_reversible is True
    assert operation.reverse() == c_ops.RenameColumnOperation(
        table=mock_table_1,
        column=mock_column_1,
        new_name=mock_column_1.name,
    )


def test_add_constraint_operation_init_reverse():
    operation = c_ops.AddConstraintOperation(
        table=mock_table_1,
        column=mock_column_1,
        constraint=c.PrimaryKeyConstraint(mock_table_1.name, mock_column_1.name),
    )
    assert operation.operation_type == OperationType.CONSTRAINT_CHANGE
    assert operation.table == mock_table_1
    assert operation.column == mock_column_1
    assert operation.constraint == c.PrimaryKeyConstraint(
        mock_table_1.name, mock_column_1.name
    )
    assert operation.is_reversible is True

    reverse_operation = operation.reverse()
    assert reverse_operation == c_ops.DeleteConstraintOperation(
        table=operation.table,
        column=operation.column,
        constraint=operation.constraint,
    )


def test_delete_constraint_operation_init_reverse():
    operation = c_ops.DeleteConstraintOperation(
        table=mock_table_1,
        column=mock_column_1,
        constraint=c.PrimaryKeyConstraint(mock_table_1.name, mock_column_1.name),
    )
    assert operation.operation_type == OperationType.CONSTRAINT_CHANGE
    assert operation.table == mock_table_1
    assert operation.column == mock_column_1
    assert operation.constraint == c.PrimaryKeyConstraint(
        mock_table_1.name, mock_column_1.name
    )
    assert operation.is_reversible is True
    reverse_operation = operation.reverse()
    assert reverse_operation == c_ops.AddConstraintOperation(
        table=operation.table,
        column=operation.column,
        constraint=operation.constraint,
    )


def test_change_data_type_operation_init_reverse():
    operation = c_ops.ChangeDataTypeOperation(
        table=mock_table_1,
        column=mock_column_1,
        new_dtype=mock_column_2.data_type,
    )
    assert operation.operation_type == OperationType.DTYPE_CHANGE
    assert operation.table == mock_table_1
    assert operation.column == mock_column_1
    assert operation.new_dtype == mock_column_2.data_type
    assert operation.is_reversible is True
    assert operation.reverse() == c_ops.ChangeDataTypeOperation(
        table=mock_table_1,
        column=mock_column_1,
        new_dtype=mock_column_1.data_type,
    )



from __future__ import annotations
from typing import Any, Self
from skibidi_orm.migration_engine.adapters.providers import DatabaseProvider


class BaseDbConfig:
    """
    Base class for all database configurations.
    Ensures that only one instance of all subclasses of this class can be created.
    """

    __instance: Any = None
    database_provider: DatabaseProvider

    @classmethod
    def get_instance(cls) -> Self:
        if BaseDbConfig.__instance is None:
            raise ReferenceError("Instance does not exist")
        return BaseDbConfig.__instance

    def __init_subclass__(cls) -> None:
        """
        Ensures that only one instance of all subclasses of this class can be created.
        """
        cls.__new__ = BaseDbConfig.__modify_new(cls)

    @staticmethod
    def __modify_new(class_t: Any):
        old_new = class_t.__new__

        def wrapper(*args: Any, **kwargs: Any) -> Any:
            if BaseDbConfig.__instance is not None:
                raise RuntimeError("Only one instance of this class is allowed")
            BaseDbConfig.__instance = old_new(class_t)
            return BaseDbConfig.__instance

        return wrapper



from skibidi_orm.migration_engine.data_mutator.base_data_mutator import BaseDataMutator
from skibidi_orm.migration_engine.data_mutator.postgres_data_mutator import (
    PostgresDataMutator,
)
from skibidi_orm.migration_engine.data_mutator.sqlite3_data_mutatorr import (
    SQLite3DataMutator,
)
from skibidi_orm.migration_engine.db_config.base_config import BaseDbConfig
from skibidi_orm.migration_engine.db_config.postgres_config import PostgresConfig
from skibidi_orm.migration_engine.db_config.sqlite3_config import SQLite3Config


def get_db_mutator(db_config: BaseDbConfig) -> BaseDataMutator:
    if isinstance(db_config, SQLite3Config):
        return SQLite3DataMutator()
    elif isinstance(db_config, PostgresConfig):
        return PostgresDataMutator()

    raise NotImplementedError



import enum


class OperationType(enum.Enum):
    """Enum class for the type of operation that can be performed on a table or column."""

    CREATE = enum.auto()
    DELETE = enum.auto()
    RENAME = enum.auto()
    DTYPE_CHANGE = enum.auto()
    CONSTRAINT_CHANGE = enum.auto()



import sqlite3
from typing import Any
from skibidi_orm.migration_engine.db_config.sqlite3_config import SQLite3Config
from skibidi_orm.migration_engine.revisions.revision import Revision
from skibidi_orm.migration_engine.sql_executor.base_sql_executor import BaseSQLExecutor
from skibidi_orm.migration_engine.operations.column_operations import ColumnOperation
from skibidi_orm.migration_engine.operations.table_operations import TableOperation

from skibidi_orm.migration_engine.converters.sqlite3.all import SQLite3Converter


class SQLite3Executor(BaseSQLExecutor):
    """
    Executes SQL statements and operations on a SQLite3 database.

    This class provides methods to execute SQL statements and operations on a SQLite3 database.
    It inherits from the BaseSQLExecutor class.

    Methods:
        execute_sql: Executes a single SQL statement.
        execute_operations: Executes a list of table or column operations.

    """

    @staticmethod
    def execute_sql(sql: str):
        """
        Executes a single SQL statement.

        Args:
            sql (str): The SQL statement to be executed.

        Returns:
            None

        """
        sqlite_config = SQLite3Config.get_instance()
        with sqlite3.connect(sqlite_config.db_path) as conn:
            cursor = conn.cursor()
            commands = sql.split(";")
            for command in commands:
                if not command:
                    continue
                cursor.execute(command.strip())
            conn.commit()

    @staticmethod
    def execute_sql_query(sql: str) -> list[Any]:
        sqlite_config = SQLite3Config.get_instance()
        with sqlite3.connect(sqlite_config.db_path) as conn:
            cursor = conn.cursor()
            result = cursor.execute(sql)

        return result.fetchall()

    @staticmethod
    def save_revision(revision: Revision):
        query = SQLite3Converter.get_revision_insertion_query()
        with sqlite3.connect(
            SQLite3Config.get_instance().db_path, detect_types=sqlite3.PARSE_DECLTYPES
        ) as conn:
            cursor = conn.cursor()
            cursor.execute(query, (revision,))
            conn.commit()

    @staticmethod
    def get_all_revisions() -> list[tuple[int, Revision]]:
        query = SQLite3Converter.get_revision_data_query()
        with sqlite3.connect(
            SQLite3Config.get_instance().db_path, detect_types=sqlite3.PARSE_DECLTYPES
        ) as conn:
            cursor = conn.cursor()
            result = cursor.execute(query)
            return result.fetchall()

    @staticmethod
    def execute_operations(operations: list[TableOperation | ColumnOperation]):
        """
        Executes a list of table or column operations.

        Args:
            operations (list[TableOperation | ColumnOperation]): The list of table or column operations to be executed.

        Returns:
            None

        """
        for operation in operations:
            SQLite3Executor.execute_sql(
                SQLite3Converter.convert_operation_to_SQL(operation)
            )



from __future__ import annotations
from dataclasses import dataclass
from typing import Any, Literal, cast
import sqlite3

from skibidi_orm.migration_engine.db_config.sqlite3_config import SQLite3Config
from skibidi_orm.migration_engine.db_inspectors.base_inspector import BaseDbInspector
from skibidi_orm.migration_engine.adapters.sqlite3_typing import (
    SQLite3Typing,
)
import skibidi_orm.migration_engine.adapters.database_objects.constraints as c
from skibidi_orm.migration_engine.revisions.constants import get_revision_table_name

type SQLite3PragmaTableInfo = list[
    tuple[int, str, str, Literal[0, 1], Any, Literal[0, 1]]
]


@dataclass(frozen=True)
class PragmaForeignKeyListEntry:
    id: int
    seq: int
    table: str
    from_column: str
    to_column: str
    on_update: str
    on_delete: str
    match: str

    @classmethod
    def from_tuple(
        cls,
        values: tuple[str, str, str, str, str, str, str, str],
    ) -> PragmaForeignKeyListEntry:
        id, seq, table, from_column, to_column, on_update, on_delete, match = values
        return cls(
            int(id),
            int(seq),
            table,
            from_column,
            to_column,
            on_update,
            on_delete,
            match,
        )


type SQLite3PragmaForeignKeyList = list[PragmaForeignKeyListEntry]


class SQLite3Inspector(BaseDbInspector):
    """
    Used to get data from live SQLite3 database.
    Should only be instantiated when SQLite3 is choosen as the database.
    """

    def __init__(self) -> None:
        self.config = SQLite3Config.get_instance()

    def get_tables(
        self,
    ) -> list[SQLite3Typing.Table]:
        """
        Retrieve all tables from the database.
        """

        tables: list[SQLite3Typing.Table] = []
        tables_names = self.get_tables_names()
        foreign_keys = self.get_foreign_key_constraints()
        for table_name in tables_names:
            table_columns = self.get_table_columns(table_name)
            related_fks = {fk for fk in foreign_keys if fk.table_name == table_name}
            tables.append(
                SQLite3Typing.Table(
                    name=table_name, columns=table_columns, foreign_keys=related_fks
                )
            )

        return tables

    def get_tables_names(self) -> list[str]:
        """
        Retrieve just tables names from the database.
        """

        tables = self._sqlite_execute(
            "SELECT name FROM sqlite_master WHERE type='table';"
        )
        return [
            name for table in tables if (name := table[0]) != get_revision_table_name()
        ]

    @staticmethod
    def foreign_keys_from_pragma_entries(
        table_entry_mapping: dict[str, list[PragmaForeignKeyListEntry]]
    ) -> set[c.ForeignKeyConstraint]:
        """Creates all of the ForeignKeyConstraint objects based on the entries
        gathered from the pragma table"""

        return_value: set[c.ForeignKeyConstraint] = set()

        for table_name, pragma_entry_list in table_entry_mapping.items():
            for id in set(entry.id for entry in pragma_entry_list):
                # todo: maybe it would be quicker to create a dict id: list[Entry] first?

                corresponding_entries = [
                    entry for entry in pragma_entry_list if entry.id == id
                ]
                referenced_table = corresponding_entries[0].table

                corresponding_constraint = c.ForeignKeyConstraint(
                    table_name,
                    referenced_table,
                    {
                        entry.from_column: entry.to_column
                        for entry in corresponding_entries
                    },
                )
                return_value.add(corresponding_constraint)

        return return_value

    def get_foreign_key_constraints(self) -> set[c.ForeignKeyConstraint]:
        """Get all foreign key constraints from the database."""

        tables_names = self.get_tables_names()
        pragma_results: dict[str, list[PragmaForeignKeyListEntry]] = {
            # Map tables to their pragma entries
            table: [
                PragmaForeignKeyListEntry.from_tuple(entry) for entry in fetched_result
            ]
            for table, fetched_result in (
                (table, self._sqlite_execute(f"PRAGMA foreign_key_list({table})"))
                for table in tables_names
            )
            if fetched_result
        }
        pragma_results = pragma_results  # todo: remove

        constraints: set[c.ForeignKeyConstraint] = (
            SQLite3Inspector.foreign_keys_from_pragma_entries(pragma_results)
        )
        return constraints

    def get_table_columns(self, table_name: str) -> list[SQLite3Typing.Column]:
        columns: SQLite3PragmaTableInfo = self._sqlite_execute(
            f"PRAGMA table_info({table_name});"
        )
        """
        When given a table name, returns a list of Column objects inside it.
        """

        return [
            SQLite3Typing.Column(
                name=name,
                data_type=cast(SQLite3Typing.DataTypes, data_type),
                column_constraints=[
                    cast(c.ColumnSpecificConstraint, constraint)
                    for constraint in [
                        (
                            c.PrimaryKeyConstraint(
                                table_name=table_name, column_name=name
                            )
                            if pk
                            else None
                        ),
                        (
                            c.NotNullConstraint(table_name=table_name, column_name=name)
                            if notnull
                            else None
                        ),
                    ]
                    if constraint is not None
                ],
            )
            for _, name, data_type, notnull, _, pk in columns
        ]

    def _sqlite_execute(self, query: str):
        """
        Execute a query in the SQLite3 database, returns its result.
        """

        db_path = self.config.db_path
        with sqlite3.connect(db_path) as conn:
            cursor = conn.cursor()
            cursor.execute(query)
            conn.commit()
            data = cursor.fetchall()
            cursor.close()
        return data



from abc import ABC, abstractmethod
from skibidi_orm.migration_engine.operations.table_operations import TableOperation
from skibidi_orm.migration_engine.operations.column_operations import ColumnOperation


class IStateManager(ABC):
    """
    Interface for the StateManager object. Can be used to calculate changes between class hierarchy schematics and
    database schematics.
    """

    @abstractmethod
    def get_operations_transforming_database_schema_into_class_hierarchy_schema(
        self,
    ) -> list[TableOperation | ColumnOperation]:
        """
        Get operations needed to transform the database structure into a class hierarchy
        defined structure
        """



from pytest import raises
import pytest
from skibidi_orm.exceptions.operations import UnsupportedOperationError
from skibidi_orm.migration_engine.adapters.postgres_typing import PostgresTyping
from skibidi_orm.migration_engine.converters.postgres.columns import (
    PostgresColumnOperationConverter,
)
from skibidi_orm.migration_engine.operations.column_operations import (
    AddColumnOperation,
    ChangeDataTypeOperation,
    DeleteColumnOperation,
    RenameColumnOperation,
)
from skibidi_orm.migration_engine.adapters.database_objects.constraints import (
    CheckConstraint,
    ForeignKeyConstraint,
    NotNullConstraint,
    PrimaryKeyConstraint,
    UniqueConstraint,
)


simple_column_no_constraints = PostgresTyping.Column("name", "TEXT")
column_primary_key_unique = PostgresTyping.Column(
    "user_id",
    "INTEGER",
    column_constraints=[
        PrimaryKeyConstraint("users", "user_id"),
        UniqueConstraint("users", "user_id"),
        NotNullConstraint("users", "user_id"),
    ],
)

column_unique = PostgresTyping.Column(
    "user_id",
    "INTEGER",
    column_constraints=[
        UniqueConstraint("users", "user_id"),
    ],
)

column_check_constraint = PostgresTyping.Column(
    "age",
    "INTEGER",
    column_constraints=[CheckConstraint("users", "age", "> 18")],
)

empty_users_table = PostgresTyping.Table("users", columns=[])


@pytest.mark.parametrize(
    "column, constraints_list, expected",
    [
        (simple_column_no_constraints, [], "name TEXT"),
        (
            column_primary_key_unique,
            column_primary_key_unique.column_constraints,
            "user_id INTEGER PRIMARY KEY UNIQUE NOT NULL",
        ),
        (
            column_unique,
            column_unique.column_constraints,
            "user_id INTEGER UNIQUE",
        ),
        (
            column_check_constraint,
            column_check_constraint.column_constraints,
            "age INTEGER CHECK (age > 18)",
        ),
    ],
)
def test_convert_column_definition_to_SQL(column, constraints_list, expected):  # type: ignore
    assert (
        PostgresColumnOperationConverter.convert_column_definition_to_SQL(
            column, constraints_list  # type: ignore
        )
        == expected
    )


@pytest.mark.parametrize(
    "operation, expected",
    [
        (
            AddColumnOperation(empty_users_table, simple_column_no_constraints),
            "ALTER TABLE users ADD COLUMN name TEXT;",
        ),
        (
            AddColumnOperation(empty_users_table, column_primary_key_unique),
            "ALTER TABLE users ADD COLUMN user_id INTEGER PRIMARY KEY UNIQUE NOT NULL;",
        ),
        (
            AddColumnOperation(
                empty_users_table,
                column_unique,
                related_foreign_key=ForeignKeyConstraint(
                    empty_users_table.name, "people", {"user_id": "person_id"}
                ),
            ),
            "ALTER TABLE users ADD COLUMN user_id INTEGER UNIQUE REFERENCES people (person_id);",
        ),
        (
            AddColumnOperation(empty_users_table, column_check_constraint),
            "ALTER TABLE users ADD COLUMN age INTEGER CHECK (age > 18);",
        ),
        (
            DeleteColumnOperation(empty_users_table, simple_column_no_constraints),
            "ALTER TABLE users DROP COLUMN name;",
        ),
        (
            DeleteColumnOperation(empty_users_table, column_primary_key_unique),
            "ALTER TABLE users DROP COLUMN user_id;",
        ),
        (
            DeleteColumnOperation(empty_users_table, column_unique),
            "ALTER TABLE users DROP COLUMN user_id;",
        ),
        (
            DeleteColumnOperation(empty_users_table, column_check_constraint),
            "ALTER TABLE users DROP COLUMN age;",
        ),
        (
            RenameColumnOperation(
                empty_users_table, simple_column_no_constraints, "name2"
            ),
            "ALTER TABLE users RENAME COLUMN name TO name2;",
        ),
        (
            RenameColumnOperation(
                empty_users_table, column_primary_key_unique, "user_id2"
            ),
            "ALTER TABLE users RENAME COLUMN user_id TO user_id2;",
        ),
        (
            RenameColumnOperation(empty_users_table, column_unique, "user_id2"),
            "ALTER TABLE users RENAME COLUMN user_id TO user_id2;",
        ),
        (
            RenameColumnOperation(empty_users_table, column_check_constraint, "age2"),
            "ALTER TABLE users RENAME COLUMN age TO age2;",
        ),
    ],
)
def test_convert_column_operation_to_SQL(operation, expected):  # type: ignore
    assert (
        PostgresColumnOperationConverter.convert_column_operation_to_SQL(operation)  # type: ignore
        == expected
    )


@pytest.mark.parametrize(
    "operation, expected_error",
    [
        (
            ChangeDataTypeOperation(
                empty_users_table, simple_column_no_constraints, "INTEGER"
            ),
            UnsupportedOperationError,
        ),
        (
            ChangeDataTypeOperation(
                empty_users_table, column_primary_key_unique, "TEXT"
            ),
            UnsupportedOperationError,
        ),
        (
            ChangeDataTypeOperation(empty_users_table, column_unique, "TEXT"),
            UnsupportedOperationError,
        ),
        (
            ChangeDataTypeOperation(empty_users_table, column_check_constraint, "TEXT"),
            UnsupportedOperationError,
        ),
    ],
)
def test_convert_column_operation_to_SQL_errors(operation, expected_error):  # type: ignore
    with raises(expected_error):  # type: ignore
        PostgresColumnOperationConverter.convert_column_operation_to_SQL(operation)  # type: ignore



from skibidi_orm.migration_engine.adapters.postgres_typing import PostgresTyping
from skibidi_orm.migration_engine.converters.base.interfaces import (
    ColumnOperationSQLConverter,
)
from skibidi_orm.migration_engine.converters.postgres.constraints import (
    PostgresConstraintConverter,
)
from skibidi_orm.migration_engine.operations.column_operations import (
    AddColumnOperation,
    ColumnOperation,
    DeleteColumnOperation,
    RenameColumnOperation,
)
from skibidi_orm.migration_engine.adapters.database_objects.constraints import (
    ColumnSpecificConstraint,
)
from skibidi_orm.migration_engine.operations.operation_type import OperationType
from skibidi_orm.exceptions.operations import UnsupportedOperationError
from typing import cast


class PostgresColumnOperationConverter(ColumnOperationSQLConverter):
    """Class responsible for converting column operation objects to raw Postgres SQL strings"""

    @staticmethod
    def convert_column_operation_to_SQL(operation: ColumnOperation) -> str:
        """Convert a given column operation to a Postgres SQL string"""
        if operation.operation_type == OperationType.CREATE:
            return PostgresColumnOperationConverter.convert_add_column_operation_to_SQL(
                cast(AddColumnOperation, operation)
            )
        elif operation.operation_type == OperationType.DELETE:
            return (
                PostgresColumnOperationConverter.convert_delete_column_operation_to_SQL(
                    cast(DeleteColumnOperation, operation)
                )
            )
        elif operation.operation_type == OperationType.RENAME:
            return (
                PostgresColumnOperationConverter.convert_rename_column_operation_to_SQL(
                    cast(RenameColumnOperation, operation)
                )
            )
        raise UnsupportedOperationError(
            "The given operation is not supported in Postgres."
        )

    @staticmethod
    def convert_add_column_operation_to_SQL(operation: AddColumnOperation) -> str:
        """Convert a given add column operation to a Postgres SQL string"""
        column_definition = (
            PostgresColumnOperationConverter.convert_column_definition_to_SQL(
                operation.column, operation.column.column_constraints
            )
        )
        return_value = (
            f"ALTER TABLE {operation.table.name} ADD COLUMN {column_definition}"
        )
        if operation.related_foreign_key is not None:
            referenced_column = operation.related_foreign_key.column_mapping[
                operation.column.name
            ]
            fk_definition = f"REFERENCES {operation.related_foreign_key.referenced_table} ({referenced_column})"
            return_value += f" {fk_definition}"

        return f"{return_value};"

    @staticmethod
    def convert_delete_column_operation_to_SQL(operation: DeleteColumnOperation) -> str:
        """Convert a given delete column operation to a Postgres SQL string"""
        return (
            f"ALTER TABLE {operation.table.name} DROP COLUMN {operation.column.name};"
        )

    @staticmethod
    def convert_rename_column_operation_to_SQL(operation: RenameColumnOperation) -> str:
        """Convert a given rename column operation to a Postgres SQL string"""
        return f"ALTER TABLE {operation.table.name} RENAME COLUMN {operation.column.name} TO {operation.new_name};"

    @staticmethod
    def convert_column_definition_to_SQL(
        column: PostgresTyping.Column, constraints: list[ColumnSpecificConstraint]
    ) -> str:
        """Convert a given column definition to a Postgres SQL string"""
        return_value = f"{column.name} {column.data_type}"
        for constraint in constraints:
            return_value += (
                f" {PostgresConstraintConverter.convert_constraint_to_SQL(constraint)}"
            )
        return return_value



import pytest
import skibidi_orm.migration_engine.adapters.database_objects.constraints as c


def test_ColumnSpecificConstraint_sortable():
    try:
        sorted(
            [
                c.PrimaryKeyConstraint(
                    table_name="table_name",
                    column_name="column_name",
                ),
                c.NotNullConstraint(
                    table_name="table_name",
                    column_name="column_name",
                ),
                c.CheckConstraint(
                    table_name="table_name",
                    column_name="column_name",
                    condition="condition",
                ),
            ]
        )
    except TypeError as e:
        pytest.fail(f"TypeError: {e}")



from dotenv import load_dotenv
import os


"""Constants regarding revision managing, such as operations used to
create revision tables, etc."""

DEFAULT_REVISION_TABLE_NAME = "__revisions"


def get_revision_table_name() -> str:
    """Get the name of the table used to store revision data.
    If no name is set via an environment variable, the default name is used."""
    load_dotenv()
    return os.environ.get("__CHUMPY_REVISION_TABLE_NAME", DEFAULT_REVISION_TABLE_NAME)  # type: ignore



from typing import Literal
from skibidi_orm.migration_engine.adapters.base_adapter import (
    BaseColumn,
    BaseTable,
)

from skibidi_orm.migration_engine.adapters.database_objects.constraints import (
    Constraint,
)


class SQLite3Typing:
    """
    Class defining the typing for SQLite3 database objects.
    """

    DataTypes = Literal["TEXT", "INTEGER", "REAL", "BLOB", "NULL", "TIMESTAMP"]
    Constraints = Constraint
    Column = BaseColumn[DataTypes]
    Table = BaseTable[Column]
    tables: list[Table] = []



from pytest import raises
from skibidi_orm.exceptions.constraints import UnsupportedConstraintError
from skibidi_orm.migration_engine.converters.sqlite3.constraints import SQLite3ConstraintConverter
from skibidi_orm.migration_engine.adapters.database_objects.constraints import (
    PrimaryKeyConstraint,
    UniqueConstraint,
    ForeignKeyConstraint,
    CheckConstraint,
    DefaultConstraint,
    NotNullConstraint,
)


def test_primary_key_constraint_conversion():
    """Test the conversion of a PrimaryKeyConstraint to SQL."""
    constraint = PrimaryKeyConstraint("users", "id")
    assert (
        SQLite3ConstraintConverter.convert_constraint_to_SQL(constraint)
        == "PRIMARY KEY"
    )


def test_unique_constraint_conversion():
    """Test the conversion of a UniqueConstraint to SQL.""" ""
    constraint = UniqueConstraint("users", "email")
    assert SQLite3ConstraintConverter.convert_constraint_to_SQL(constraint) == "UNIQUE"


def test_not_null_constraint_conversion():
    """Test the conversion of a NotNullConstraint to SQL."""
    constraint = NotNullConstraint("users", "email")
    assert (
        SQLite3ConstraintConverter.convert_constraint_to_SQL(constraint) == "NOT NULL"
    )


def test_default_constraint_conversion():
    """Test the conversion of a DefaultConstraint to SQL."""
    with raises(UnsupportedConstraintError):
        SQLite3ConstraintConverter.convert_constraint_to_SQL(
            DefaultConstraint("users", "name", "Gordon")
        )


def test_single_column_foreign_key_constraint_conversion():
    """Test the conversion of a ForeignKeyConstraint with a single column to SQL."""
    constraint = ForeignKeyConstraint("users", "addresses", {"address_id": "id"})
    assert SQLite3ConstraintConverter.convert_constraint_to_SQL(constraint) == (
        "FOREIGN KEY (address_id) REFERENCES addresses (id)"
    )


def test_composite_foreign_key_constraint_conversion():
    """Test the conversion of a ForeignKeyConstraint with multiple columns to SQL.""" ""
    constraint = ForeignKeyConstraint(
        "admins", "users", {"user_id": "id", "address": "address", "group_id": "group"}
    )
    assert (
        SQLite3ConstraintConverter.convert_constraint_to_SQL(constraint)
        == "FOREIGN KEY (user_id, address, group_id) REFERENCES users (id, address, group)"
    )


def test_check_constraint_conversion():
    """Test the conversion of a CheckConstraint to SQL.""" ""
    constraint = CheckConstraint("users", "age", "> 18")
    assert (
        SQLite3ConstraintConverter.convert_constraint_to_SQL(constraint)
        == "CHECK (age > 18)"
    )



from skibidi_orm.migration_engine.db_config.sqlite3_config import SQLite3Config
from skibidi_orm.migration_engine.db_inspectors.sqlite3_inspector import (
    SQLite3Inspector,
)
from skibidi_orm.migration_engine.adapters.database_objects.constraints import (
    PrimaryKeyConstraint,
)
from skibidi_orm.migration_engine.sql_executor.sqlite3_executor import SQLite3Executor

type TestTableRow = tuple[int, str | None]


def test_execute_sql(make_database: str):
    SQLite3Config(make_database)
    SQLite3Executor.execute_sql(
        "CREATE TABLE test_table (id INTEGER PRIMARY KEY, name TEXT);"
    )
    table = SQLite3Inspector().get_tables()[0]
    assert table.name == "test_table"
    assert table.columns[0].name == "id"
    assert table.columns[0].data_type == "INTEGER"
    assert table.columns[0].column_constraints == [
        PrimaryKeyConstraint("test_table", "id")
    ]
    assert table.columns[1].name == "name"
    assert table.columns[1].data_type == "TEXT"
    assert table.columns[1].column_constraints == []


def test_execute_query(make_database: str):
    SQLite3Config(make_database)
    SQLite3Executor.execute_sql(
        """
        CREATE TABLE test_table (id INTEGER PRIMARY KEY, name TEXT);
        INSERT INTO test_table (id, name) VALUES (1, 'test1');
        INSERT INTO test_table (id) VALUES (2);
        """
    )
    table_contents: list[TestTableRow] = SQLite3Executor.execute_sql_query(
        "SELECT * FROM test_table;"
    )

    assert table_contents == [(1, "test1"), (2, None)]



from pytest import raises
from skibidi_orm.exceptions.operations import UnsupportedOperationError
from skibidi_orm.migration_engine.adapters.sqlite3_typing import SQLite3Typing
from skibidi_orm.migration_engine.converters.sqlite3.columns import (
    SQLite3ColumnOperationConverter,
)
from skibidi_orm.migration_engine.operations.column_operations import (
    AddColumnOperation,
    ChangeDataTypeOperation,
    DeleteColumnOperation,
    RenameColumnOperation,
)
from skibidi_orm.migration_engine.adapters.database_objects.constraints import (
    CheckConstraint,
    ForeignKeyConstraint,
    NotNullConstraint,
    PrimaryKeyConstraint,
    UniqueConstraint,
)


simple_column_no_constraints = SQLite3Typing.Column("name", "TEXT")
column_primary_key_unique = SQLite3Typing.Column(
    "user_id",
    "INTEGER",
    column_constraints=[
        PrimaryKeyConstraint("users", "user_id"),
        UniqueConstraint("users", "user_id"),
        NotNullConstraint("users", "user_id"),
    ],
)

column_unique = SQLite3Typing.Column(
    "user_id",
    "INTEGER",
    column_constraints=[
        UniqueConstraint("users", "user_id"),
    ],
)

column_check_constraint = SQLite3Typing.Column(
    "age",
    "INTEGER",
    column_constraints=[CheckConstraint("users", "age", "> 18")],
)

empty_users_table = SQLite3Typing.Table("users", columns=[])


def test_simple_column_definition():
    assert (
        SQLite3ColumnOperationConverter.convert_column_definition_to_SQL(
            simple_column_no_constraints, list()
        )
        == "name TEXT"
    )


def test_column_definition_with_primary_key_unique_nn():
    assert (
        SQLite3ColumnOperationConverter.convert_column_definition_to_SQL(
            column_primary_key_unique, column_primary_key_unique.column_constraints
        )
        == "user_id INTEGER PRIMARY KEY UNIQUE NOT NULL"
    )


def test_column_definition_unique():
    assert (
        SQLite3ColumnOperationConverter.convert_column_definition_to_SQL(
            column_unique, column_unique.column_constraints
        )
        == "user_id INTEGER UNIQUE"
    )


def test_column_definition_with_check_constraint():
    assert (
        SQLite3ColumnOperationConverter.convert_column_definition_to_SQL(
            column_check_constraint, column_check_constraint.column_constraints
        )
        == "age INTEGER CHECK (age > 18)"
    )


def test_add_simple_column():
    operation = AddColumnOperation(empty_users_table, simple_column_no_constraints)
    assert (
        SQLite3ColumnOperationConverter.convert_column_operation_to_SQL(operation)
        == "ALTER TABLE users ADD COLUMN name TEXT;"
    )


def test_add_column_with_primary_key_and_unique():
    operation = AddColumnOperation(empty_users_table, column_primary_key_unique)
    assert (
        SQLite3ColumnOperationConverter.convert_column_operation_to_SQL(operation)
        == "ALTER TABLE users ADD COLUMN user_id INTEGER PRIMARY KEY UNIQUE NOT NULL;"
    )


def test_add_column_with_foreign_key_and_unique():
    operation = AddColumnOperation(
        empty_users_table,
        column_unique,
        related_foreign_key=ForeignKeyConstraint(
            empty_users_table.name, "people", {"user_id": "person_id"}
        ),
    )
    assert (
        SQLite3ColumnOperationConverter.convert_column_operation_to_SQL(operation)
        == "ALTER TABLE users ADD COLUMN user_id INTEGER UNIQUE REFERENCES people (person_id);"
    )


def test_add_column_with_check_constraint():
    operation = AddColumnOperation(empty_users_table, column_check_constraint)
    assert (
        SQLite3ColumnOperationConverter.convert_column_operation_to_SQL(operation)
        == "ALTER TABLE users ADD COLUMN age INTEGER CHECK (age > 18);"
    )


def test_drop_simple_column():
    operation = DeleteColumnOperation(empty_users_table, simple_column_no_constraints)
    assert (
        SQLite3ColumnOperationConverter.convert_column_operation_to_SQL(operation)
        == "ALTER TABLE users DROP COLUMN name;"
    )


def test_drop_column_with_primary_key_and_unique():
    operation = DeleteColumnOperation(empty_users_table, column_primary_key_unique)
    assert (
        SQLite3ColumnOperationConverter.convert_column_operation_to_SQL(operation)
        == "ALTER TABLE users DROP COLUMN user_id;"
    )


def test_drop_column_with_foreign_key_and_unique():
    operation = DeleteColumnOperation(empty_users_table, column_unique)
    assert (
        SQLite3ColumnOperationConverter.convert_column_operation_to_SQL(operation)
        == "ALTER TABLE users DROP COLUMN user_id;"
    )


def test_drop_column_with_check_constraint():
    operation = DeleteColumnOperation(empty_users_table, column_check_constraint)
    assert (
        SQLite3ColumnOperationConverter.convert_column_operation_to_SQL(operation)
        == "ALTER TABLE users DROP COLUMN age;"
    )


def test_rename_simple_column():
    operation = RenameColumnOperation(
        empty_users_table, simple_column_no_constraints, "name2"
    )
    assert (
        SQLite3ColumnOperationConverter.convert_column_operation_to_SQL(operation)
        == "ALTER TABLE users RENAME COLUMN name TO name2;"
    )


def test_rename_column_with_primary_key_and_unique():
    operation = RenameColumnOperation(
        empty_users_table, column_primary_key_unique, "user_id2"
    )
    assert (
        SQLite3ColumnOperationConverter.convert_column_operation_to_SQL(operation)
        == "ALTER TABLE users RENAME COLUMN user_id TO user_id2;"
    )


def test_rename_column_with_foreign_key_and_unique():
    operation = RenameColumnOperation(empty_users_table, column_unique, "user_id2")
    assert (
        SQLite3ColumnOperationConverter.convert_column_operation_to_SQL(operation)
        == "ALTER TABLE users RENAME COLUMN user_id TO user_id2;"
    )


def test_rename_column_with_check_constraint():
    operation = RenameColumnOperation(
        empty_users_table, column_check_constraint, "age2"
    )
    assert (
        SQLite3ColumnOperationConverter.convert_column_operation_to_SQL(operation)
        == "ALTER TABLE users RENAME COLUMN age TO age2;"
    )


def test_change_data_type_simple_column():
    operation = ChangeDataTypeOperation(
        empty_users_table, simple_column_no_constraints, "INTEGER"
    )
    with raises(UnsupportedOperationError):
        SQLite3ColumnOperationConverter.convert_column_operation_to_SQL(operation)


def test_change_data_type_primary_key_unique_nn():
    operation = ChangeDataTypeOperation(
        empty_users_table, column_primary_key_unique, "TEXT"
    )
    with raises(UnsupportedOperationError):
        SQLite3ColumnOperationConverter.convert_column_operation_to_SQL(operation)


def test_change_data_type_column_with_foreign_key_and_unique():
    operation = ChangeDataTypeOperation(empty_users_table, column_unique, "TEXT")

    with raises(UnsupportedOperationError):
        SQLite3ColumnOperationConverter.convert_column_operation_to_SQL(operation)


def test_change_data_type_column_with_check_constraint():
    operation = ChangeDataTypeOperation(
        empty_users_table, column_check_constraint, "TEXT"
    )
    with raises(UnsupportedOperationError):
        SQLite3ColumnOperationConverter.convert_column_operation_to_SQL(operation)



import pathlib
from skibidi_orm.migration_engine.db_config.sqlite3_config import SQLite3Config
from skibidi_orm.migration_engine.db_inspectors.sqlite3_inspector import (
    SQLite3Inspector,
)
from skibidi_orm.migration_engine.studio.utils.get_db_inspector import get_db_inspector
import py  # type: ignore


def write_temp_schema_file(file_path: str, file_content: str):
    with open(file_path, "w") as f:
        f.write(file_content)


def test_get_db_inspector_sqlite3(tmp_path: pathlib.Path):
    p = tmp_path.joinpath("schema_test_get_db_inspector_sqlite3.py")
    p.write_text(
        """from skibidi_orm.migration_engine.db_config.sqlite3_config import SQLite3Config
SQLite3Config(db_path="some/path/to/db.db")"""
    )
    db_inspector = get_db_inspector(
        db_config=SQLite3Config(db_path="some/path/to/db.db")
    )
    assert isinstance(db_inspector, SQLite3Inspector)



import glob
import os
from skibidi_orm.exceptions.cli_exceptions import MultipleSchemaFilesError
import importlib.util
import sys

GLOB_ALL_SCHEMAS = "**/**/schema.py"


def find_schema_file():
    """
    Find the schema file in the project.
    If no schema file is found, raise FileNotFoundError.
    If multiple schema files are found, raise MultipleSchemaFilesError.
    """
    files = set(glob.glob(GLOB_ALL_SCHEMAS, recursive=True))
    if len(files) == 0:
        raise FileNotFoundError("No schema file found.")
    if len(files) > 1:
        raise MultipleSchemaFilesError("Multiple schema files found.")
    return files.pop()


def find_schema_file_with_a_given_path(path: str):
    """
    Find the schema file under the given path and return it in a .

    If no schema file is found, raise FileNotFoundError.
    """

    if os.path.exists(path):
        return path
    else:
        raise FileNotFoundError("No schema file found.")


def load_schema_from_path(path: str):
    """
    Load the schema file from the given path into the workspace.

    If schema file is not found, raise FileNotFoundError, if it cannot be imported, raise ImportError.
    It changes the current working directory to the specified path. Then it loads the schema file, and as a result,
    the absolute path of the database is relative to the schema file. After loading the schema file, it changes the
    current working directory back to the original one.
    """

    path = os.path.abspath(path)

    tmp_cwd = os.getcwd()
    os.chdir(os.path.dirname(path))

    path = find_schema_file_with_a_given_path(path)

    try:
        spec = importlib.util.spec_from_file_location("schema", path)
        module = importlib.util.module_from_spec(spec)  # type: ignore

        sys.modules["schema"] = module
        spec.loader.exec_module(module)  # type: ignore

    except ImportError:
        raise ImportError("Error while importing the schema file.")

    os.chdir(tmp_cwd)



from skibidi_orm.migration_engine.db_config.base_config import BaseDbConfig
from skibidi_orm.migration_engine.db_config.postgres_config import PostgresConfig
from skibidi_orm.migration_engine.db_config.sqlite3_config import SQLite3Config
from skibidi_orm.migration_engine.db_inspectors.base_inspector import BaseDbInspector
from skibidi_orm.migration_engine.db_inspectors.postgres_inspector import (
    PostgresInspector,
)
from skibidi_orm.migration_engine.db_inspectors.sqlite3_inspector import (
    SQLite3Inspector,
)


def get_db_inspector(db_config: BaseDbConfig) -> BaseDbInspector:
    if isinstance(db_config, SQLite3Config):
        return SQLite3Inspector()
    elif isinstance(db_config, PostgresConfig):
        return PostgresInspector()

    raise NotImplementedError



from datetime import datetime
from dataclasses import dataclass, field
from typing import Any
import sqlite3
import pickle

from skibidi_orm.migration_engine.adapters.base_adapter import BaseColumn, BaseTable
from skibidi_orm.migration_engine.adapters.providers import DatabaseProvider


@dataclass
class Revision:
    """Class representing a revision of the schema at a given time.
    After serializing it, it is stored in the revision table.
    Every database provider has to take care of adapting and converting it
    before inserting it into the table.

    Attributes:
        timestamp (datetime): Timestamp of the creation.

        description (str): Migration message provided by the user when migrating.

        schema_repr (str): String representation of the schema.

        provider (str): Database provider.

        tables (list[BaseTable]): List of tables in the schema.
    """

    timestamp: datetime = field(init=False, default_factory=lambda: datetime.now())
    description: str
    schema_repr: str
    provider: DatabaseProvider
    tables: list[BaseTable[BaseColumn[Any]]]

    def __conform__(self, protocol: Any):
        """This method is used to serialize the object before inserting it into the database.
        Every database provider has to take care of adapting and converting it."""
        if protocol is sqlite3.PrepareProtocol:
            return pickle.dumps(self)

    @staticmethod
    def deserialize(data: bytes):
        """Deserializes the data from the database."""
        return pickle.loads(data)


# register the converter function which deserializes the object
sqlite3.register_converter("REVISION", Revision.deserialize)



from skibidi_orm.migration_engine.adapters.base_adapter import BaseAdapter

from skibidi_orm.migration_engine.operations.column_operations import ColumnOperation
from skibidi_orm.migration_engine.operations.table_operations import TableOperation


# All classes need to inherit from this class if they participate in a migration
class MigrationElement:
    """
    Base class for all migration elements. Every class that wants to participate in a migration should inherit from this class.
    """

    adapter: BaseAdapter
    operations: list[ColumnOperation | TableOperation] = []

    def migrate(self, preview: bool = False):
        """
        This will execute the migration for all the migration elements.

        """
        table = MigrationElement.__subclasses__()

        for cls in table:
            table_instance = cls()
            table_instance.adapter.execute_migration(preview)



"""
Notice about postgres constraints and types:

1. SERIAL/BIGSERIAL:
    - expands to INTEGER or BIGINT, respectively with a default set to nextval('table_name_column_name_seq'::regclass)
    - Using SERIAL or BIGSERIAL automatically makes the column NOT NULL.
2. Aliases:
    - There are many aliases such as VARCHAR for CHARACTER VARYING, TIMESTAMPTZ for TIMESTAMP WITH TIME ZONE, etc.
    - Postgres internally stores expanded versions, so PostgresInspector should return the expanded version of the type.
3. CHECK constraints:
    - CHECK constraints are internally stored as strings for every column referenced in the CHECK constraint.
    - Thus, PostgresInspector returns CHECK constaint for every column that is referenced in the CHECK constraint.
4. DEFAULT constraints:
    - Some default constaints are stored as is, e.g.
        `my_col INTEGER DEFAULT 1`, will store `1` as the default value.
    - Other default constaints are stored with type casting, e.g.
        `my_col TEXT DEFAULT 'default'`, will store `'default'::text` as the default value.
    - Other default constaints store functions names, e.g.
        `my_col DATE DEFAULT CURRENT_DATE`, will store `CURRENT_DATE` as the default value.
    - PostgresInspector returns the default values as stored internally by Postgres.
"""

import pytest
from skibidi_orm.migration_engine.adapters.postgres_typing import PostgresTyping
from skibidi_orm.migration_engine.db_config.postgres_config import PostgresConfig
from skibidi_orm.migration_engine.db_inspectors.postgres_inspector import (
    PostgresInspector,
)
from ..conftest import postgres_db_fixture
from ..sql_data import PostgresTablesData, SQLite3TablesData
import skibidi_orm.migration_engine.adapters.database_objects.constraints as c


@pytest.mark.parametrize(
    "query, expected_tables_len, expected_tables_list",
    [
        ([SQLite3TablesData.sql_table1], 1, ["table1"]),
        (
            [SQLite3TablesData.sql_table1, SQLite3TablesData.sql_table2],
            2,
            ["table1", "table2"],
        ),
        (
            [SQLite3TablesData.sql_table_primary_key_not_null],
            1,
            ["table_primary_key_not_null"],
        ),
    ],
)
def test_get_tables_names(query, expected_tables_list, expected_tables_len):  # type: ignore
    @postgres_db_fixture(
        db_name="postgres",
        db_user="admin",
        db_password="admin",
        db_host="0.0.0.0",
        db_port=5432,
        queries=query,  # type: ignore
    )
    def test_fn(config: PostgresConfig, inspector: PostgresInspector):
        tables = inspector.get_tables_names()
        assert len(tables) == expected_tables_len
        assert sorted(expected_tables_list) == sorted(tables)  # type: ignore

    test_fn()


def test__get_table_columns_names():
    @postgres_db_fixture(
        db_name="postgres",
        db_user="admin",
        db_password="admin",
        db_host="0.0.0.0",
        db_port=5432,
        queries=[PostgresTablesData.SQL_TABLE_MANY_COLUMNS1],
    )
    def test_fn(config: PostgresConfig, inspector: PostgresInspector):
        columns_names = inspector._get_table_columns_names("table_many_columns1")  # type: ignore
        assert len(columns_names) == 7
        assert sorted(columns_names) == sorted(
            [
                "primary_key",
                "text_not_null",
                "text_nullabe",
                "integer_not_null",
                "integer_nullable",
                "date_not_null",
                "date_nullable",
            ]
        )

    test_fn()


@pytest.mark.parametrize(
    "query, table_name, column_name, expected_data_type",
    [
        (  # 0
            [PostgresTablesData.SQL_TABLE_MANY_COLUMNS1],
            "table_many_columns1",
            "primary_key",
            "INTEGER",
        ),
        (  # 1
            [PostgresTablesData.SQL_TABLE_MANY_COLUMNS1],
            "table_many_columns1",
            "text_not_null",
            "TEXT",
        ),
        (  # 2
            [PostgresTablesData.SQL_TABLE_MANY_COLUMNS1],
            "table_many_columns1",
            "text_nullabe",
            "TEXT",
        ),
        (  # 3
            [PostgresTablesData.SQL_TABLE_MANY_COLUMNS1],
            "table_many_columns1",
            "integer_not_null",
            "INTEGER",
        ),
        (  # 4
            [PostgresTablesData.SQL_TABLE_MANY_COLUMNS1],
            "table_many_columns1",
            "integer_nullable",
            "INTEGER",
        ),
        (  # 5
            [PostgresTablesData.SQL_TABLE_MANY_COLUMNS1],
            "table_many_columns1",
            "date_not_null",
            "DATE",
        ),
        (  # 6
            [PostgresTablesData.SQL_TABLE_MANY_COLUMNS1],
            "table_many_columns1",
            "date_nullable",
            "DATE",
        ),
        (  # 7
            [PostgresTablesData.SQL_TABLE_DIFFICULT_TYPES],
            "table_difficult_types",
            "big_primary_key_alias",
            "BIGINT",
        ),
        (  # 8
            [PostgresTablesData.SQL_TABLE_DIFFICULT_TYPES],
            "table_difficult_types",
            "bit_varying_alias",
            "BIT VARYING",
        ),
        (  # 9
            [PostgresTablesData.SQL_TABLE_DIFFICULT_TYPES],
            "table_difficult_types",
            "bit_varying_alias_arg",
            "BIT VARYING",
        ),
        (  # 10
            [PostgresTablesData.SQL_TABLE_DIFFICULT_TYPES],
            "table_difficult_types",
            "double_precision",
            "DOUBLE PRECISION",
        ),
        (  # 11
            [PostgresTablesData.SQL_TABLE_DIFFICULT_TYPES],
            "table_difficult_types",
            "timestamp",
            "TIMESTAMP WITHOUT TIME ZONE",
        ),
        (  # 12
            [PostgresTablesData.SQL_TABLE_DIFFICULT_TYPES],
            "table_difficult_types",
            "timestamp_arg",
            "TIMESTAMP WITHOUT TIME ZONE",
        ),
        (  # 13
            [PostgresTablesData.SQL_TABLE_DIFFICULT_TYPES],
            "table_difficult_types",
            "timestamp_without_time_zone",
            "TIMESTAMP WITHOUT TIME ZONE",
        ),
        (  # 14
            [PostgresTablesData.SQL_TABLE_DIFFICULT_TYPES],
            "table_difficult_types",
            "timestamp_without_time_zone_arg",
            "TIMESTAMP WITHOUT TIME ZONE",
        ),
        (  # 15
            [PostgresTablesData.SQL_TABLE_DIFFICULT_TYPES],
            "table_difficult_types",
            "timestamp_with_time_zone_alias",
            "TIMESTAMP WITH TIME ZONE",
        ),
        (  # 16
            [PostgresTablesData.SQL_TABLE_DIFFICULT_TYPES],
            "table_difficult_types",
            "timestamp_with_time_zone_arg_alias",
            "TIMESTAMP WITH TIME ZONE",
        ),
        (  # 17
            [PostgresTablesData.SQL_TABLE_DIFFICULT_TYPES],
            "table_difficult_types",
            "charachter_varying",
            "CHARACTER VARYING",
        ),
        (  # 18
            [PostgresTablesData.SQL_TABLE_DIFFICULT_TYPES],
            "table_difficult_types",
            "charachter_varying_arg",
            "CHARACTER VARYING",
        ),
        (  # 19
            [PostgresTablesData.SQL_TABLE_DIFFICULT_TYPES],
            "table_difficult_types",
            "charachter_varying_alias",
            "CHARACTER VARYING",
        ),
        (  # 20
            [PostgresTablesData.SQL_TABLE_DIFFICULT_TYPES],
            "table_difficult_types",
            "charachter_varying_alias_args",
            "CHARACTER VARYING",
        ),
    ],
)
def test__get_column_data_type(query, table_name, column_name, expected_data_type):  # type: ignore
    @postgres_db_fixture(
        db_name="postgres",
        db_user="admin",
        db_password="admin",
        db_host="0.0.0.0",
        db_port=5432,
        queries=query,  # type: ignore
    )
    def test_fn(config: PostgresConfig, inspector: PostgresInspector):
        data_type = inspector._get_column_data_type(table_name, column_name)  # type: ignore
        assert data_type == expected_data_type

    test_fn()


@pytest.mark.parametrize(
    "query, table_name, column_name, expected_nullable",
    [
        (  # 0
            [PostgresTablesData.SQL_TABLE_MANY_COLUMNS1],
            "table_many_columns1",
            "text_not_null",
            False,
        ),
        (  # 1
            [PostgresTablesData.SQL_TABLE_MANY_COLUMNS1],
            "table_many_columns1",
            "text_nullabe",
            True,
        ),
        (  # 2
            [PostgresTablesData.SQL_TABLE_MANY_COLUMNS1],
            "table_many_columns1",
            "integer_not_null",
            False,
        ),
        (  # 3
            [PostgresTablesData.SQL_TABLE_MANY_COLUMNS1],
            "table_many_columns1",
            "integer_nullable",
            True,
        ),
        (  # 4
            [PostgresTablesData.SQL_TABLE_MANY_COLUMNS1],
            "table_many_columns1",
            "date_not_null",
            False,
        ),
        (  # 5
            [PostgresTablesData.SQL_TABLE_MANY_COLUMNS1],
            "table_many_columns1",
            "date_nullable",
            True,
        ),
        (  # 6
            [PostgresTablesData.SQL_TABLE_MANY_COLUMNS1],
            "table_many_columns1",
            "primary_key",
            False,
        ),
    ],
)
def test__is_column_nullable(query, table_name, column_name, expected_nullable):  # type: ignore
    @postgres_db_fixture(
        db_name="postgres",
        db_user="admin",
        db_password="admin",
        db_host="0.0.0.0",
        db_port=5432,
        queries=query,  # type: ignore
    )
    def test_fn(config: PostgresConfig, inspector: PostgresInspector):
        assert inspector._is_column_nullable(table_name, column_name) is expected_nullable  # type: ignore

    test_fn()


@pytest.mark.parametrize(
    "query, table_name, column_name, expected_constraints",
    [
        (  # 0
            [PostgresTablesData.SQL_TABLE_DIFFERECT_CONSTRAINTS],
            "table_different_constraints",
            "primary_key",
            [
                c.NotNullConstraint("table_different_constraints", "primary_key"),
                c.PrimaryKeyConstraint("table_different_constraints", "primary_key"),
            ],
        ),
        (  # 1
            [PostgresTablesData.SQL_TABLE_DIFFERECT_CONSTRAINTS],
            "table_different_constraints",
            "text_nullable",
            [],
        ),
        (  # 2
            [PostgresTablesData.SQL_TABLE_DIFFERECT_CONSTRAINTS],
            "table_different_constraints",
            "integer_not_nullable",
            [
                c.NotNullConstraint(
                    "table_different_constraints", "integer_not_nullable"
                ),
                c.CheckConstraint(
                    "table_different_constraints",
                    "integer_not_nullable",
                    "((integer_not_nullable > 100))",
                ),
            ],
        ),
        (  # 3
            [PostgresTablesData.SQL_TABLE_DIFFERECT_CONSTRAINTS],
            "table_different_constraints",
            "unique_column",
            [c.UniqueConstraint("table_different_constraints", "unique_column")],
        ),
        (  # 4
            [PostgresTablesData.SQL_TABLE_DIFFERECT_CONSTRAINTS],
            "table_different_constraints",
            "unique_not_nullable",
            [
                c.UniqueConstraint(
                    "table_different_constraints", "unique_not_nullable"
                ),
                c.NotNullConstraint(
                    "table_different_constraints", "unique_not_nullable"
                ),
            ],
        ),
        (  # 5
            [PostgresTablesData.SQL_TABLE_DIFFERECT_CONSTRAINTS],
            "table_different_constraints",
            "check_other_column",
            [],
        ),
        (  # 6
            [PostgresTablesData.SQL_TABLE_DIFFERECT_CONSTRAINTS],
            "table_different_constraints",
            "default_column",
            [c.DefaultConstraint("table_different_constraints", "default_column", "1")],
        ),
        (  # 7
            [PostgresTablesData.SQL_TABLE_DIFFERECT_CONSTRAINTS],
            "table_different_constraints",
            "not_null_unique_check_column",
            [
                c.NotNullConstraint(
                    "table_different_constraints", "not_null_unique_check_column"
                ),
                c.UniqueConstraint(
                    "table_different_constraints", "not_null_unique_check_column"
                ),
                c.CheckConstraint(
                    "table_different_constraints",
                    "not_null_unique_check_column",
                    "((not_null_unique_check_column > 100))",
                ),
            ],
        ),
        (  # 8
            PostgresTablesData.SQL_TABLE_SIMPLE_FOREIGN_KEYS,
            "authors",
            "id",
            [
                c.NotNullConstraint("authors", "id"),
                c.PrimaryKeyConstraint("authors", "id"),
                c.DefaultConstraint(
                    "authors", "id", "nextval('authors_id_seq'::regclass)"
                ),
            ],
        ),
        (  # 9
            PostgresTablesData.SQL_TABLE_SIMPLE_FOREIGN_KEYS,
            "books",
            "author_id_foreign_key",
            [],
        ),
        (  # 10
            [PostgresTablesData.SQL_TABLE_DIFFERECT_CONSTRAINTS],
            "table_different_constraints",
            "not_null_unique",
            [
                c.NotNullConstraint("table_different_constraints", "not_null_unique"),
                c.UniqueConstraint("table_different_constraints", "not_null_unique"),
            ],
        ),
        (  # 11
            [PostgresTablesData.SQL_TABLE_DIFFERECT_CONSTRAINTS],
            "table_different_constraints",
            "not_null_default",
            [
                c.NotNullConstraint("table_different_constraints", "not_null_default"),
                c.DefaultConstraint(
                    "table_different_constraints", "not_null_default", "1"
                ),
            ],
        ),
        (  # 12
            [PostgresTablesData.SQL_TABLE_DIFFERECT_CONSTRAINTS],
            "table_different_constraints",
            "not_null_unique_check_default",
            [
                c.NotNullConstraint(
                    "table_different_constraints", "not_null_unique_check_default"
                ),
                c.UniqueConstraint(
                    "table_different_constraints", "not_null_unique_check_default"
                ),
                c.DefaultConstraint(
                    "table_different_constraints", "not_null_unique_check_default", "1"
                ),
                c.CheckConstraint(
                    "table_different_constraints",
                    "not_null_unique_check_default",
                    "((not_null_unique_check_default > 100))",
                ),
            ],
        ),
        (  # 13
            [PostgresTablesData.SQL_TABLE_SERIALS],
            "table_serials",
            "serial_pk",
            [
                c.DefaultConstraint(
                    "table_serials",
                    "serial_pk",
                    "nextval('table_serials_serial_pk_seq'::regclass)",
                ),
                c.NotNullConstraint("table_serials", "serial_pk"),
                c.PrimaryKeyConstraint("table_serials", "serial_pk"),
            ],
        ),
        (  # 14
            [PostgresTablesData.SQL_TABLE_SERIALS],
            "table_serials",
            "bigserial",
            [
                c.DefaultConstraint(
                    "table_serials",
                    "bigserial",
                    "nextval('table_serials_bigserial_seq'::regclass)",
                ),
                c.NotNullConstraint("table_serials", "bigserial"),
            ],
        ),
        (  # 15
            [PostgresTablesData.SQL_TABLE_DEFAULTS],
            "table_defaults",
            "integer_default",
            [
                c.DefaultConstraint(
                    "table_defaults",
                    "integer_default",
                    "1",
                ),
            ],
        ),
        (  # 16
            [PostgresTablesData.SQL_TABLE_DEFAULTS],
            "table_defaults",
            "text_default",
            [
                c.DefaultConstraint(
                    "table_defaults",
                    "text_default",
                    "'default'::text",
                ),
            ],
        ),
        (  # 17
            [PostgresTablesData.SQL_TABLE_DEFAULTS],
            "table_defaults",
            "date_default",
            [
                c.DefaultConstraint(
                    "table_defaults",
                    "date_default",
                    "CURRENT_DATE",
                ),
            ],
        ),
        (  # 18
            [PostgresTablesData.SQL_TABLE_DEFAULTS],
            "table_defaults",
            "timestamp_default",
            [
                c.DefaultConstraint(
                    "table_defaults",
                    "timestamp_default",
                    "CURRENT_TIMESTAMP",
                ),
            ],
        ),
        (  # 19
            [PostgresTablesData.SQL_TABLE_DEFAULTS],
            "table_defaults",
            "timestamp_default_lowercase",
            [
                c.DefaultConstraint(
                    "table_defaults",
                    "timestamp_default_lowercase",
                    "CURRENT_TIMESTAMP",
                ),
            ],
        ),
    ],
)
def test__get_column_constraints(query, table_name, column_name, expected_constraints):  # type: ignore
    @postgres_db_fixture(
        db_name="postgres",
        db_user="admin",
        db_password="admin",
        db_host="0.0.0.0",
        db_port=5432,
        queries=query,  # type: ignore
    )
    def test_fn(config: PostgresConfig, inspector: PostgresInspector):
        constraints = inspector._get_column_constraints(table_name, column_name)  # type: ignore
        assert sorted(constraints) == sorted(expected_constraints)  # type: ignore

    test_fn()


@pytest.mark.parametrize(
    "query, table_name, expected_columns",
    [
        (  # 0
            [PostgresTablesData.SQL_TABLE_MANY_COLUMNS1],
            "table_many_columns1",
            [
                PostgresTyping.Column(
                    name="primary_key",
                    data_type="INTEGER",
                    column_constraints=[
                        c.NotNullConstraint("table_many_columns1", "primary_key"),
                        c.PrimaryKeyConstraint("table_many_columns1", "primary_key"),
                    ],
                ),
                PostgresTyping.Column(
                    name="text_not_null",
                    data_type="TEXT",
                    column_constraints=[
                        c.NotNullConstraint("table_many_columns1", "text_not_null"),
                    ],
                ),
                PostgresTyping.Column(
                    name="text_nullabe",
                    data_type="TEXT",
                    column_constraints=[],
                ),
                PostgresTyping.Column(
                    name="integer_not_null",
                    data_type="INTEGER",
                    column_constraints=[
                        c.NotNullConstraint("table_many_columns1", "integer_not_null"),
                    ],
                ),
                PostgresTyping.Column(
                    name="integer_nullable",
                    data_type="INTEGER",
                    column_constraints=[],
                ),
                PostgresTyping.Column(
                    name="date_not_null",
                    data_type="DATE",
                    column_constraints=[
                        c.NotNullConstraint("table_many_columns1", "date_not_null"),
                    ],
                ),
                PostgresTyping.Column(
                    name="date_nullable",
                    data_type="DATE",
                    column_constraints=[],
                ),
            ],
        ),
        (  # 1
            [PostgresTablesData.SQL_TABLE_DIFFICULT_TYPES],
            "table_difficult_types",
            [
                PostgresTyping.Column(
                    name="big_primary_key_alias",
                    data_type="BIGINT",
                    column_constraints=[
                        c.DefaultConstraint(
                            "table_difficult_types",
                            "big_primary_key_alias",
                            "nextval('table_difficult_types_big_primary_key_alias_seq'::regclass)",
                        ),
                        c.NotNullConstraint(
                            "table_difficult_types", "big_primary_key_alias"
                        ),
                        c.PrimaryKeyConstraint(
                            "table_difficult_types",
                            "big_primary_key_alias",
                        ),
                    ],
                ),
                PostgresTyping.Column(
                    name="bit_varying_alias",
                    data_type="BIT VARYING",
                    column_constraints=[],
                ),
                PostgresTyping.Column(
                    name="bit_varying_alias_arg",
                    data_type="BIT VARYING",
                    column_constraints=[],
                ),
                PostgresTyping.Column(
                    name="double_precision",
                    data_type="DOUBLE PRECISION",
                    column_constraints=[],
                ),
                PostgresTyping.Column(
                    name="timestamp",
                    data_type="TIMESTAMP WITHOUT TIME ZONE",
                    column_constraints=[],
                ),
                PostgresTyping.Column(
                    name="timestamp_arg",
                    data_type="TIMESTAMP WITHOUT TIME ZONE",
                    column_constraints=[],
                ),
                PostgresTyping.Column(
                    name="timestamp_without_time_zone",
                    data_type="TIMESTAMP WITHOUT TIME ZONE",
                    column_constraints=[],
                ),
                PostgresTyping.Column(
                    name="timestamp_without_time_zone_arg",
                    data_type="TIMESTAMP WITHOUT TIME ZONE",
                    column_constraints=[],
                ),
                PostgresTyping.Column(
                    name="timestamp_with_time_zone_alias",
                    data_type="TIMESTAMP WITH TIME ZONE",
                    column_constraints=[],
                ),
                PostgresTyping.Column(
                    name="timestamp_with_time_zone_arg_alias",
                    data_type="TIMESTAMP WITH TIME ZONE",
                    column_constraints=[],
                ),
                PostgresTyping.Column(
                    name="charachter_varying",
                    data_type="CHARACTER VARYING",
                    column_constraints=[],
                ),
                PostgresTyping.Column(
                    name="charachter_varying_arg",
                    data_type="CHARACTER VARYING",
                    column_constraints=[],
                ),
                PostgresTyping.Column(
                    name="charachter_varying_alias",
                    data_type="CHARACTER VARYING",
                    column_constraints=[],
                ),
                PostgresTyping.Column(
                    name="charachter_varying_alias_args",
                    data_type="CHARACTER VARYING",
                    column_constraints=[],
                ),
            ],
        ),
        (  # 2
            [PostgresTablesData.SQL_TABLE_DIFFERECT_CONSTRAINTS],
            "table_different_constraints",
            [
                PostgresTyping.Column(
                    name="primary_key",
                    data_type="INTEGER",
                    column_constraints=[
                        c.NotNullConstraint(
                            "table_different_constraints", "primary_key"
                        ),
                        c.PrimaryKeyConstraint(
                            "table_different_constraints", "primary_key"
                        ),
                    ],
                ),
                PostgresTyping.Column(
                    name="integer_not_nullable",
                    data_type="INTEGER",
                    column_constraints=[
                        c.NotNullConstraint(
                            "table_different_constraints", "integer_not_nullable"
                        ),
                        c.CheckConstraint(
                            "table_different_constraints",
                            "integer_not_nullable",
                            "((integer_not_nullable > 100))",
                        ),
                    ],
                ),
                PostgresTyping.Column(
                    name="text_nullable",
                    data_type="TEXT",
                    column_constraints=[],
                ),
                PostgresTyping.Column(
                    name="unique_column",
                    data_type="INTEGER",
                    column_constraints=[
                        c.UniqueConstraint(
                            "table_different_constraints", "unique_column"
                        ),
                    ],
                ),
                PostgresTyping.Column(
                    name="unique_not_nullable",
                    data_type="INTEGER",
                    column_constraints=[
                        c.UniqueConstraint(
                            "table_different_constraints", "unique_not_nullable"
                        ),
                        c.NotNullConstraint(
                            "table_different_constraints", "unique_not_nullable"
                        ),
                    ],
                ),
                PostgresTyping.Column(
                    name="check_other_column",
                    data_type="INTEGER",
                    column_constraints=[],
                ),
                PostgresTyping.Column(
                    name="default_column",
                    data_type="INTEGER",
                    column_constraints=[
                        c.DefaultConstraint(
                            "table_different_constraints", "default_column", "1"
                        ),
                    ],
                ),
                PostgresTyping.Column(
                    name="not_null_unique_check_column",
                    data_type="INTEGER",
                    column_constraints=[
                        c.NotNullConstraint(
                            "table_different_constraints",
                            "not_null_unique_check_column",
                        ),
                        c.UniqueConstraint(
                            "table_different_constraints",
                            "not_null_unique_check_column",
                        ),
                        c.CheckConstraint(
                            "table_different_constraints",
                            "not_null_unique_check_column",
                            "((not_null_unique_check_column > 100))",
                        ),
                    ],
                ),
                PostgresTyping.Column(
                    name="not_null_unique",
                    data_type="INTEGER",
                    column_constraints=[
                        c.NotNullConstraint(
                            "table_different_constraints", "not_null_unique"
                        ),
                        c.UniqueConstraint(
                            "table_different_constraints", "not_null_unique"
                        ),
                    ],
                ),
                PostgresTyping.Column(
                    name="not_null_default",
                    data_type="INTEGER",
                    column_constraints=[
                        c.NotNullConstraint(
                            "table_different_constraints", "not_null_default"
                        ),
                        c.DefaultConstraint(
                            "table_different_constraints", "not_null_default", "1"
                        ),
                    ],
                ),
                PostgresTyping.Column(
                    name="not_null_unique_check_default",
                    data_type="INTEGER",
                    column_constraints=[
                        c.NotNullConstraint(
                            "table_different_constraints",
                            "not_null_unique_check_default",
                        ),
                        c.UniqueConstraint(
                            "table_different_constraints",
                            "not_null_unique_check_default",
                        ),
                        c.DefaultConstraint(
                            "table_different_constraints",
                            "not_null_unique_check_default",
                            "1",
                        ),
                        c.CheckConstraint(
                            "table_different_constraints",
                            "not_null_unique_check_default",
                            "((not_null_unique_check_default > 100))",
                        ),
                    ],
                ),
            ],
        ),
    ],
)
def test_get_table_columns(query, table_name, expected_columns):  # type: ignore
    @postgres_db_fixture(
        db_name="postgres",
        db_user="admin",
        db_password="admin",
        db_host="0.0.0.0",
        db_port=5432,
        queries=query,  # type: ignore
    )
    def test_fn(config: PostgresConfig, inspector: PostgresInspector):
        columns = inspector.get_table_columns(table_name)  # type: ignore
        for column in columns:
            column.column_constraints = sorted(column.column_constraints)
        for column in expected_columns:  # type: ignore
            column.column_constraints = sorted(column.column_constraints)  # type: ignore
        assert sorted(columns) == sorted(expected_columns)  # type: ignore

    test_fn()


@pytest.mark.parametrize(
    "query, table_name, expected_foreign_keys",
    [
        (  # 0
            PostgresTablesData.SQL_TABLE_FKS_MORE_COMPLEX,
            "departments",
            set(),  # type: ignore
        ),
        (  # 1
            PostgresTablesData.SQL_TABLE_FKS_MORE_COMPLEX,
            "employees",
            set(
                [
                    c.ForeignKeyConstraint(
                        table_name="employees",
                        referenced_table="departments",
                        column_mapping={
                            "department_id": "department_id",
                        },
                    )
                ]
            ),
        ),
        (  # 2
            PostgresTablesData.SQL_TABLE_FKS_MORE_COMPLEX,
            "projects",
            set(
                [
                    c.ForeignKeyConstraint(
                        table_name="projects",
                        referenced_table="employees",
                        column_mapping={
                            "lead_employee_id": "employee_id",
                        },
                    ),
                    c.ForeignKeyConstraint(
                        table_name="projects",
                        referenced_table="departments",
                        column_mapping={
                            "support_department_id": "department_id",
                        },
                    ),
                ]
            ),
        ),
    ],
)
def test__get_foreign_keys(query, table_name, expected_foreign_keys):  # type: ignore
    @postgres_db_fixture(
        db_name="postgres",
        db_user="admin",
        db_password="admin",
        db_host="0.0.0.0",
        db_port=5432,
        queries=query,  # type: ignore
    )
    def test_fn(config: PostgresConfig, inspector: PostgresInspector):
        foreign_keys = inspector._get_foreign_keys(table_name)  # type: ignore
        assert foreign_keys == expected_foreign_keys

    test_fn()



from typing import Literal
from skibidi_orm.migration_engine.adapters.base_adapter import (
    BaseColumn,
    BaseTable,
)

from skibidi_orm.migration_engine.adapters.database_objects.constraints import (
    Constraint,
)


class PostgresTyping:

    DataTypes = Literal[
        "BIGINT",
        "BIGSERIAL",
        "BIT",
        "BIT VARYING",
        "BOOLEAN",
        "BOX",
        "BYTEA",
        "CHARACTER",
        "CHARACTER VARYING",
        "CIDR",
        "CIRCLE",
        "DATE",
        "DOUBLE PRECISION",
        "INET",
        "INTEGER",
        "INTERVAL",
        "JSON",
        "JSONB",
        "LINE",
        "LSEG",
        "MACADDR",
        "MACADDR8",
        "MONEY",
        "NUMERIC",
        "PATH",
        "PG_LSN",
        "POINT",
        "POLYGON",
        "REAL",
        "SMALLINT",
        "SMALLSERIAL",
        "SERIAL",
        "TEXT",
        "TIME",
        "TIME WITH TIME ZONE",
        "TIMESTAMP",
        "TIMESTAMP WITH TIME ZONE",
        "TIMESTAMP WITHOUT TIME ZONE",
        "TSQUERY",
        "TSVECTOR",
        "TXID_SNAPSHOT",
        "UUID",
        "XML",
    ]
    Constraints = Constraint
    Column = BaseColumn[DataTypes]
    Table = BaseTable[Column]



from typing import Any

from psycopg2 import ProgrammingError
from skibidi_orm.migration_engine.data_mutator.base_data_mutator import (
    BaseDataMutator,
    DeleteRowPk,
    InsertRowColumn,
)
from skibidi_orm.migration_engine.db_config.postgres_config import PostgresConfig


class PostgresDataMutator(BaseDataMutator):
    def insert_row(self, table_name: str, row: list[InsertRowColumn]) -> None:
        """Insert a row in the table."""
        raise NotImplementedError

    def delete_row(self, table_name: str, pks: list[DeleteRowPk]) -> None:
        """Delete a row in the table. Row identified by primary key subset."""
        raise NotImplementedError

    def raw_query(self, query: str) -> list[Any]:
        """Execute a raw sql query in the database."""
        config = PostgresConfig.get_instance()
        with config.connection.cursor() as cursor:
            cursor.execute(query)
            try:
                return cursor.fetchall()
            except ProgrammingError:
                return []

    def get_rows(self, table_name: str, limit: int = 100, offset: int = 0) -> list[Any]:
        """Get paginated rows from the table."""
        config = PostgresConfig.get_instance()
        with config.connection.cursor() as cursor:
            cursor.execute(f"SELECT * FROM {table_name} LIMIT {limit} OFFSET {offset}")
            return cursor.fetchall()



from skibidi_orm.migration_engine.adapters.providers import DatabaseProvider
from skibidi_orm.migration_engine.db_config.base_config import (
    BaseDbConfig,
)
import os


# TODO: make a dataclass out of config classes
class SQLite3Config(BaseDbConfig):
    """
    Configuration class for SQLite3 database.
    Instantiating it means choosing SQLite3 as the database.
    """

    database_provider = DatabaseProvider.SQLITE3

    def __init__(self, db_path: str):
        self.__db_path = os.path.abspath(db_path)

    @property
    def db_path(self) -> str:
        return self.__db_path



from skibidi_orm.migration_engine.converters.base.interfaces import (
    ColumnOperationSQLConverter,
    ConstraintSQLConverter,
    SQLConverter,
    TableOperationSQLConverter,
)
from skibidi_orm.migration_engine.converters.postgres.columns import (
    PostgresColumnOperationConverter,
)
from skibidi_orm.migration_engine.converters.postgres.constraints import (
    PostgresConstraintConverter,
)
from skibidi_orm.migration_engine.converters.postgres.tables import (
    PostgresTableOperationConverter,
)


class PostgresConverter(SQLConverter):
    """Class reponsible for converting operation and constraint
    objects to Postgres SQL strings"""

    @staticmethod
    def get_constraint_converter() -> type[ConstraintSQLConverter]:
        return PostgresConstraintConverter

    @staticmethod
    def get_table_operation_converter() -> type[TableOperationSQLConverter]:
        return PostgresTableOperationConverter

    @staticmethod
    def get_column_operation_converter() -> type[ColumnOperationSQLConverter]:
        return PostgresColumnOperationConverter



import pytest
from skibidi_orm.exceptions.config_exceptions import DbConnectionError
from skibidi_orm.migration_engine.db_config.postgres_config import PostgresConfig
from psycopg2.extensions import connection as Connection


def test_create_connection_valid():
    """
    Should connect to running Postgres database.
    Make sure to have a running Postgres database.
    """
    config = PostgresConfig(
        db_name="postgres",
        db_user="admin",
        db_password="admin",
        db_host="0.0.0.0",
        db_port=5432,
    )
    assert config.connection is not None
    assert isinstance(config.connection, Connection)
    assert config.connection.closed == 0
    config.connection.close()
    assert config.connection.closed == 1


@pytest.mark.slow
@pytest.mark.parametrize(
    "db_name, db_user, db_password, db_host, db_port",
    [
        ("invalid", "admin", "admin", "admin", 5432),
        ("postgres", "invalid", "admin", "admin", 5432),
        ("postgres", "admin", "invalid", "admin", 5432),
        ("postgres", "admin", "admin", "invalid", 5432),
        ("postgres", "admin", "admin", "admin", 1337),
    ],
)
def test_should_raise_DbConnectionError_on_invalid_connection(
    db_name, db_user, db_password, db_host, db_port  # type: ignore
):
    with pytest.raises(DbConnectionError):
        PostgresConfig(
            db_name=db_name,  # type: ignore
            db_user=db_user,  # type: ignore
            db_password=db_password,  # type: ignore
            db_host=db_host,  # type: ignore
            db_port=db_port,  # type: ignore
        )



from pytest import raises
import pytest
from skibidi_orm.exceptions.constraints import UnsupportedConstraintError
from skibidi_orm.migration_engine.converters.sqlite3.constraints import (
    SQLite3ConstraintConverter,
)
from skibidi_orm.migration_engine.adapters.database_objects.constraints import (
    PrimaryKeyConstraint,
    UniqueConstraint,
    ForeignKeyConstraint,
    CheckConstraint,
    DefaultConstraint,
    NotNullConstraint,
)


@pytest.mark.parametrize(
    "constraint, expected",
    [
        (PrimaryKeyConstraint("users", "id"), "PRIMARY KEY"),
        (UniqueConstraint("users", "email"), "UNIQUE"),
        (NotNullConstraint("users", "email"), "NOT NULL"),
        (
            ForeignKeyConstraint("users", "addresses", {"address_id": "id"}),
            "FOREIGN KEY (address_id) REFERENCES addresses (id)",
        ),
        (
            ForeignKeyConstraint(
                "admins",
                "users",
                {"user_id": "id", "address": "address", "group_id": "group"},
            ),
            "FOREIGN KEY (user_id, address, group_id) REFERENCES users (id, address, group)",
        ),
        (CheckConstraint("users", "age", "> 18"), "CHECK (age > 18)"),
    ],
)
def test_convert_constraint_to_SQL(constraint, expected):  # type: ignore
    """Test the conversion of various constraints to SQL."""
    assert SQLite3ConstraintConverter.convert_constraint_to_SQL(constraint) == expected  # type: ignore


@pytest.mark.parametrize(
    "constraint, expected_error",
    [
        (DefaultConstraint("users", "name", "Gordon"), UnsupportedConstraintError),
    ],
)
def test_convert_constraint_to_SQL_error(constraint, expected_error):  # type: ignore
    """Test the conversion of a DefaultConstraint to SQL."""
    with raises(expected_error):  # type: ignore
        SQLite3ConstraintConverter.convert_constraint_to_SQL(constraint)  # type: ignore



import psycopg2
from skibidi_orm.exceptions.config_exceptions import DbConnectionError
from skibidi_orm.migration_engine.adapters.providers import DatabaseProvider
from skibidi_orm.migration_engine.db_config.base_config import (
    BaseDbConfig,
)
from psycopg2.extensions import connection as Connection


class PostgresConfig(BaseDbConfig):

    database_provider = DatabaseProvider.POSTGRESQL

    """
    Configuration class for Postgres database.
    Instantiating it means choosing Postgres as the database.
    """

    def __init__(
        self, db_name: str, db_user: str, db_password: str, db_host: str, db_port: int
    ):
        self.__db_name = db_name
        self.__db_user = db_user
        self.__db_password = db_password
        self.__db_host = db_host
        self.__db_port = db_port
        self.__connection = self.__create_connection()

    @property
    def db_name(self) -> str:
        return self.__db_name

    @property
    def db_user(self) -> str:
        return self.__db_user

    @property
    def db_password(self) -> str:
        return self.__db_password

    @property
    def db_host(self) -> str:
        return self.__db_host

    @property
    def db_port(self) -> int:
        return self.__db_port

    @property
    def connection(self) -> Connection:
        return self.__connection

    def __create_connection(self) -> Connection:
        try:
            connection = psycopg2.connect(
                user=self.db_user,
                password=self.db_password,
                host=self.db_host,
                port=self.db_port,
                database=self.db_name,
            )
        except psycopg2.OperationalError as e:
            raise DbConnectionError(e)

        return connection



from __future__ import annotations
from abc import ABC, abstractmethod
from skibidi_orm.migration_engine.operations.operation_type import OperationType
from skibidi_orm.migration_engine.adapters.database_objects.constraints import (
    Constraint,
    ForeignKeyConstraint,
)
from skibidi_orm.exceptions.operations import IrreversibleOperationError
from skibidi_orm.migration_engine.adapters.base_adapter import BaseTable, BaseColumn
from typing import Any
from dataclasses import dataclass, field
from typing import Optional


@dataclass(frozen=True)
class ColumnOperation(ABC):
    """Base class for column operations"""

    is_reversible: bool = field(init=False)
    operation_type: OperationType = field(init=False)
    table: BaseTable[BaseColumn[Any]]
    column: BaseColumn[Any]

    @abstractmethod
    def reverse(self) -> ColumnOperation:
        """Generates a table operation that reverses
        the calling instance if possible. Else, throws an exception"""

    def get_related_foreign_keys(self) -> list[ForeignKeyConstraint]:
        """Returns the foreign keys that are related to the column operation.
        If the column is already instantiated and is a part of a foreign key,
        it has to be taken into acount when removing and adding it back to the schema.
        """
        return [
            fk
            for fk in self.table.foreign_keys
            if self.column.name in fk.column_mapping
        ]

    @abstractmethod
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(table={self.table}, is_reversible={self.is_reversible})"


@dataclass(frozen=True)
class AddColumnOperation(ColumnOperation):
    """Class for adding a column"""

    operation_type: OperationType = field(init=False, default=OperationType.CREATE)
    is_reversible: bool = field(init=False, default=True)

    # if the column to be added references another table, it has to be set here
    related_foreign_key: Optional[ForeignKeyConstraint] = field(default=None)

    def reverse(self) -> ColumnOperation:
        return DeleteColumnOperation(table=self.table, column=self.column)

    def __post_init__(self):
        self.validate_related_foreign_key()

    def validate_related_foreign_key(self) -> None:
        """This method validates the related_foreign_key attribute of the AddColumnOperation.
        The field is supposed to serve as a way to add a non-composite foreign key to the table
        as a column is being added. Trying to add a composite foreign key or one that does not have
        the added column as the origin column will result in an error."""
        if self.related_foreign_key is not None:
            if len(self.related_foreign_key.column_mapping.values()) != 1:
                raise ValueError(
                    "The related_foreign_key should have exactly one column mapping, with the key being the column that's added to the table"
                )
            if self.column.name not in self.related_foreign_key.column_mapping:
                raise ValueError(
                    "The only possible key value for the column mapping in an AddColumnOperation is the added column name"
                )
            if self.table.name != self.related_foreign_key.table_name:
                raise ValueError(
                    f"""Foreign key constraint initialized with invalid source table name: {self.related_foreign_key.table_name}.
                    Expected: {self.table.name}"""
                )

    def __str__(self) -> str:
        return f"Add Column {self.column.name} to Table {self.table.name}"

    def __repr__(self) -> str:
        return super().__repr__()


@dataclass(frozen=True)
class DeleteColumnOperation(ColumnOperation):
    """Class for deleting a column"""

    operation_type: OperationType = field(init=False, default=OperationType.DELETE)
    is_reversible: bool = field(init=False, default=False)

    def reverse(self) -> ColumnOperation:
        raise IrreversibleOperationError(
            f"Reversing a {self.__class__.__name__} is currently not supported."
        )

    def __str__(self) -> str:
        return f"Delete Column {self.column.name} from Table {self.table.name}"

    def __repr__(self) -> str:
        return super().__repr__()


@dataclass(frozen=True)
class RenameColumnOperation(ColumnOperation):
    """Class for renaming a column"""

    operation_type: OperationType = field(init=False, default=OperationType.RENAME)
    is_reversible: bool = field(init=False, default=True)

    new_name: str

    def reverse(self) -> ColumnOperation:
        return RenameColumnOperation(
            table=self.table,
            column=self.column,
            new_name=self.column.name,
        )

    def __str__(self) -> str:
        return f"Rename Column {self.column.name} to {self.new_name} in Table {self.table.name}"

    def __repr__(self) -> str:
        return super().__repr__()


@dataclass(frozen=True)
class AddConstraintOperation(ColumnOperation):
    """Class for adding a constraint on a column"""

    operation_type: OperationType = field(
        init=False, default=OperationType.CONSTRAINT_CHANGE
    )
    is_reversible: bool = field(init=False, default=True)

    constraint: Constraint

    def reverse(self) -> ColumnOperation:
        return DeleteConstraintOperation(
            table=self.table,
            column=self.column,
            constraint=self.constraint,
        )

    def __str__(self) -> str:
        return f"Add Constraint {self.constraint.constraint_type.value} to Column {self.column.name} in Table {self.table.name}"

    def __repr__(self) -> str:
        return super().__repr__()


@dataclass(frozen=True)
class DeleteConstraintOperation(ColumnOperation):
    """Class for deleting a constraint on a column"""

    operation_type: OperationType = field(
        init=False, default=OperationType.CONSTRAINT_CHANGE
    )
    is_reversible: bool = field(init=False, default=True)

    constraint: Constraint

    def reverse(self) -> ColumnOperation:
        return AddConstraintOperation(
            table=self.table,
            column=self.column,
            constraint=self.constraint,
        )

    def __str__(self) -> str:
        return f"Delete Constraint {self.constraint.constraint_type.value} from Column {self.column.name} in Table {self.table.name}"

    def __repr__(self) -> str:
        return super().__repr__()


@dataclass(frozen=True)
class ChangeDataTypeOperation(ColumnOperation):
    """Class for changing the data type of a column"""

    operation_type: OperationType = field(
        init=False, default=OperationType.DTYPE_CHANGE
    )
    is_reversible: bool = field(init=False, default=True)
    new_dtype: str

    def reverse(self) -> ColumnOperation:
        return ChangeDataTypeOperation(
            table=self.table,
            column=self.column,
            new_dtype=self.column.data_type,
        )

    def __str__(self) -> str:
        return f"Change Data Type of Column {self.column.name} in Table {self.table.name} to {self.new_dtype}"

    def __repr__(self) -> str:
        return super().__repr__()



from skibidi_orm.migration_engine.converters.base.interfaces import SQLQueryConverter
from skibidi_orm.migration_engine.revisions.constants import get_revision_table_name


class SQLite3QueryConverter(SQLQueryConverter):
    """Class responsible for converting query objects to SQLite3 SQL strings"""

    @staticmethod
    def get_revision_data_query() -> str:
        return f"SELECT rowid, * FROM {get_revision_table_name()};"

    @staticmethod
    def get_table_clearing_query() -> str:
        return f"""
            PRAGMA writable_schema = 1;
            DELETE FROM sqlite_master WHERE type IN ('table', 'index', 'trigger') AND name != '{get_revision_table_name()}';
            PRAGMA writable_schema = 0;
            COMMIT;
            VACUUM;
            PRAGMA INTEGRITY_CHECK;
            """



import os
from typing import Literal, TextIO, Any
from pathlib import Path
import black
from skibidi_orm.migration_engine.adapters.base_adapter import BaseColumn, BaseTable
from skibidi_orm.migration_engine.adapters.database_objects.migration_element import (
    MigrationElement,
)
from skibidi_orm.migration_engine.db_config.base_config import BaseDbConfig
from skibidi_orm.migration_engine.db_config.sqlite3_config import SQLite3Config
from skibidi_orm.migration_engine.operations import table_operations as TO
from skibidi_orm.migration_engine.operations import column_operations as CO

type Typing = Literal["SQLite3Typing"]


def create_migration_file(migration_element: MigrationElement):
    """
    Inspect a migration element and create a Python file that can be executed to apply the migration.
    """
    migration_file_path = os.path.join(*[os.getcwd(), "migrations", "migration.py"])

    if not os.path.isdir(os.path.join(*[os.getcwd(), "migrations"])):
        os.mkdir(os.path.join(*[os.getcwd(), "migrations"]))

    config = BaseDbConfig.get_instance()

    migration_element.migrate(preview=True)

    operations = migration_element.operations

    if isinstance(config, SQLite3Config):
        executor = "SQLite3Executor"
        converter = "SQLite3Converter"
        typing = "SQLite3Typing"
    else:
        raise NotImplementedError

    operation_methods = extract_operation_methods(operations, typing)

    with open(migration_file_path, "w") as file:
        write_configuration(file)

        file.write("def migrate():\n")
        for method in operation_methods:

            if "TableOperation" in method:
                file.write(
                    f"\n\t{executor}.execute_sql(\n\t\t{converter}.convert_table_operation_to_SQL(\n\t\t\t{method}\n\t\t)\n\t)"
                )
            elif "ColumnOperation" in method:
                file.write(
                    f"\n\t{executor}.execute_sql({converter}.convert_column_operation_to_SQL({method}))"
                )
        if not operation_methods:
            file.write("\n\t# No operations to execute\n")
            file.write("\n\tpass")
        file.write("\n\n")

        file.write("if __name__ == '__main__':\n")
        file.write("\tmigrate()\n")

    black.format_file_in_place(
        Path(migration_file_path),
        fast=True,
        mode=black.FileMode(),
        write_back=black.WriteBack.YES,
    )


def extract_operation_methods(
    operations: list[CO.ColumnOperation | TO.TableOperation],
    typing: Typing,
) -> list[str]:
    """
    From a list of operations, extract proper strings of methods that can be executed in the migration file.
    """

    operation_methods: list[str] = []

    for operation in operations:
        if isinstance(operation, TO.TableOperation):
            operation_methods.append(parse_table_operation(operation, typing))
        else:
            operation_methods.append(parse_column_operation(operation, typing))

    return operation_methods


def write_configuration(file: TextIO):
    """
    Write the configuration for the migration file.
    """

    config = BaseDbConfig.get_instance()

    if isinstance(config, SQLite3Config):
        write_sqlite3_configuration(file)
    else:
        raise NotImplementedError

    file.write("# This is an auto-generated migration file.\n")
    file.write(
        "# Executing this file will result in executing the proposed migration.\n\n"
    )


def write_sqlite3_configuration(file: TextIO):
    """
    Write database specific configuration for SQLite3.
    """

    config = SQLite3Config.get_instance()

    file.write(
        """
from skibidi_orm.migration_engine.db_config.sqlite3_config import SQLite3Config
from skibidi_orm.migration_engine.sql_executor.sqlite3_executor import SQLite3Executor
from skibidi_orm.migration_engine.converters.sqlite3.all import SQLite3Converter
from skibidi_orm.migration_engine.adapters.database_objects import constraints as c
from skibidi_orm.migration_engine.adapters.sqlite3_typing import SQLite3Typing
from skibidi_orm.migration_engine.operations import table_operations as TO
from skibidi_orm.migration_engine.operations import column_operations as CO
"""
    )

    file.write("# Database: SQLite3\n")
    file.write(f"# Path: {config.db_path}\n\n")

    file.write("\n\n\n")
    file.write(f'SQLite3Config("{config.db_path}")')
    file.write("\n\n\n")


def parse_column_operation(operation: CO.ColumnOperation, typing: Typing) -> str:
    """
    Parse a ColumnOperation object to get a creatable string
    """
    operation_str = ""

    column = operation.column

    column_str = return_column_string(column, typing)

    table_str = return_table_string(operation.table, typing)

    if isinstance(operation, CO.RenameColumnOperation):
        operation_str = f"CO.{operation.__class__.__name__}(table={table_str}, column={column_str}, new_name='{operation.new_name}')"
    else:
        operation_str = (
            f"CO.{operation.__class__.__name__}(table={table_str}, column={column_str})"
        )

    return operation_str


def parse_table_operation(operation: TO.TableOperation, typing: Typing) -> str:
    """
    Parse a TableOperation object to get a creatable string
    """
    operation_str = ""
    table_str = ""

    table_str = return_table_string(operation.table, typing)

    if isinstance(operation, TO.RenameTableOperation):
        operation_str = f"TO.{operation.__class__.__name__}(table={table_str}, new_name='{operation.new_name}')"
    else:
        operation_str = f"TO.{operation.__class__.__name__}(table={table_str})"

    return operation_str


def return_column_string(column: BaseColumn[Any], typing: Typing) -> str:
    """
    Return a string that can construct a column object.
    """

    constraints_str = return_constraint_list_string(column, typing)

    return f"{typing}.Column(name='{column.name}', data_type='{column.data_type}', column_constraints={constraints_str},),"


def return_constraint_list_string(column: BaseColumn[Any], typing: Typing) -> str:
    """
    Return a string representation of a list of constraints.
    """
    constraints_str = "["

    for constraint in column.column_constraints:
        constraints_str += f"c.{constraint}, "
    constraints_str = constraints_str[:-2] + "]"

    return constraints_str


def return_table_string(table: BaseTable[BaseColumn[Any]], typing: Typing) -> str:
    """
    Return a string that can create a table object.
    """
    columns_str = "["
    for column in table.columns:
        columns_str += return_column_string(column, typing)
    columns_str += "]"

    return f"{typing}.Table(name='{table.name}', columns={columns_str})"



import os
from typing import cast
from fastapi import FastAPI, Body
from fastapi.responses import FileResponse
from fastapi.staticfiles import StaticFiles
import uvicorn
from skibidi_orm.migration_engine.data_mutator.base_data_mutator import (
    BaseDataMutator,
    DeleteRowPk,
    InsertRowColumn,
)
from fastapi.middleware.cors import CORSMiddleware
from skibidi_orm.migration_engine.db_inspectors.base_inspector import BaseDbInspector
from skibidi_orm.migration_engine.studio.utils.db_config_dynamic_import import (
    db_config_dynamic_import,
)
from skibidi_orm.migration_engine.studio.utils.get_db_inspector import get_db_inspector
from skibidi_orm.migration_engine.studio.utils.get_db_seeder import get_db_mutator

app = FastAPI()
db_inspector: BaseDbInspector = cast(
    BaseDbInspector, {}
)  # to add db_inspector to modules namespace
db_mutator: BaseDataMutator = cast(
    BaseDataMutator, {}
)  # to add db_seeder to modules namespace

app.add_middleware(
    CORSMiddleware,
    allow_origins=["http://localhost:5173", "http://localhost:8000"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

app.mount(
    "/assets",
    StaticFiles(
        directory=os.path.join(
            os.path.dirname(__file__), "./skibidi-orm-studio-frontend/dist/assets"
        )
    ),
    name="static",
)


def run_server(schema_file: str):
    global db_inspector
    global db_mutator
    db_config = db_config_dynamic_import(schema_file_path=schema_file)
    db_inspector = get_db_inspector(db_config=db_config)
    db_mutator = get_db_mutator(db_config=db_config)
    uvicorn.run(app, host="0.0.0.0", port=8000)


@app.get("/")
def read_index():
    return FileResponse(
        os.path.join(
            os.path.dirname(__file__), "./skibidi-orm-studio-frontend/dist/index.html"
        )
    )


@app.get("/db")
def get_db():
    """Get the tables in the database."""
    tables = db_inspector.get_tables()
    return {"tables": tables}


@app.post("/db/{table_name}/row")
def insert_row(table_name: str, row: list[InsertRowColumn] = Body(embed=True)):
    """Insert a row in the table."""
    if table_name not in db_inspector.get_tables_names():
        return {"message": "Table does not exist."}

    db_mutator.insert_row(table_name=table_name, row=row)
    return {"message": "Row inserted successfully."}


@app.post("/db/{table_name}/row/delete")
def delete_row(table_name: str, pks: list[DeleteRowPk] = Body(embed=True)):
    """Delete a row in the table. Row identified by primary key subset."""
    if table_name not in db_inspector.get_tables_names():
        return {"message": "Table does not exist."}

    db_mutator.delete_row(table_name=table_name, pks=pks)
    return {"message": "Row deleted successfully."}


@app.post("/db/query")
def query_table(query: str = Body(embed=True)):
    """Execute a raw sql query in the database."""
    results = db_mutator.raw_query(query)
    return results


@app.get("/db/{table_name}/rows")
def get_rows(table_name: str, offset: int = 0, limit: int = 100):
    """Get paginated rows from the table."""
    if table_name not in db_inspector.get_tables_names():
        return {"message": "Table does not exist."}

    rows = db_mutator.get_rows(table_name, offset=offset, limit=limit)
    return rows



from __future__ import annotations
from skibidi_orm.migration_engine.adapters.base_adapter import BaseAdapter
from skibidi_orm.migration_engine.adapters.database_objects.migration_element import (
    MigrationElement,
)
from skibidi_orm.migration_engine.adapters.postgres_typing import PostgresTyping
from skibidi_orm.migration_engine.db_inspectors.postgres_inspector import (
    PostgresInspector,
)
from skibidi_orm.migration_engine.state_manager.state_manager import StateManager


class PostgresAdapter(BaseAdapter):

    tables: list[PostgresTyping.Table]

    def __init__(self) -> None:
        self.tables = []

    def create_table(self, table: PostgresTyping.Table):
        """Informs the adapter about Table creation."""

        self.tables.append(table)

    def execute_migration(self, preview: bool = False):
        """Execute the migration process on a full adapter."""

        self.inspector = PostgresInspector()

        db_tables = self.inspector.get_tables()

        state_manager = StateManager[PostgresTyping.Table](
            db_tables=db_tables,
            schema_tables=self.tables,
        )

        MigrationElement.operations = (
            state_manager.get_operations_transforming_database_schema_into_class_hierarchy_schema()
        )

        # if not preview:
        #     SQLite3Executor.execute_operations(MigrationElement.operations)



from skibidi_orm.migration_engine.adapters.sqlite3_typing import (
    SQLite3Typing,
)
from skibidi_orm.migration_engine.adapters.sqlite3_adapter import SQLite3Adapter
from skibidi_orm.migration_engine.db_config.sqlite3_config import SQLite3Config
from skibidi_orm.migration_engine.sql_executor.sqlite3_executor import SQLite3Executor

from skibidi_orm.migration_engine.adapters.database_objects import constraints as c
from skibidi_orm.migration_engine.operations.column_operations import (
    AddColumnOperation,
    DeleteColumnOperation,
)
from skibidi_orm.migration_engine.operations.table_operations import (
    CreateTableOperation,
    DeleteTableOperation,
    RenameTableOperation,
)

from skibidi_orm.migration_engine.adapters.database_objects.migration_element import (
    MigrationElement,
)


from pathlib import Path
import pytest
import sqlite3
from ..sql_data import SQLite3TablesData


def execute_sqlite3_commands(db_path: str, commands: list[str]):
    """Executes the given commands in a SQLite DB living under db_path"""
    with sqlite3.connect(db_path) as conn:
        cursor = conn.cursor()
        for command in commands:
            cursor.execute(command)
        conn.commit()
        cursor.close()


@pytest.fixture
def tmp_database(request: pytest.FixtureRequest, tmp_path: Path):
    """Fixture for creating a temporary database and executing sqlite3 code. Yields the
    path of the db file."""
    sql_commands = request.param
    (tmp_file := tmp_path.joinpath("tmp_db.db")).touch()
    execute_sqlite3_commands(
        str(tmp_file),
        sql_commands,
    )
    yield str(tmp_file)


@pytest.fixture
def mock_execute_operations(monkeypatch: pytest.MonkeyPatch):
    # Define a function that does nothing
    def do_nothing(operations):  # type: ignore
        print("do_nothing")
        pass

    # Patch the execute_operations method with the do_nothing function
    monkeypatch.setattr(
        SQLite3Executor,
        "execute_operations",
        do_nothing,  # type: ignore
    )


@pytest.mark.parametrize(
    "tmp_database",
    [[SQLite3TablesData.sql_table_user, SQLite3TablesData.sql_table_post]],
    indirect=True,
)
def test_rename_table_operation_needed(
    tmp_database: str, capfd: pytest.CaptureFixture[str]
):
    from skibidi_orm.migration_engine.adapters.database_objects.migration_element import (
        MigrationElement,
    )

    class Table(MigrationElement):

        def __init__(self) -> None:
            self.adapter = SQLite3Adapter()

            models = Table.__subclasses__()
            if self.__class__ == Table:
                for cls in models:
                    self.adapter.create_table(cls.__dict__["table"])

    class User(Table):  # type: ignore
        columns = [
            SQLite3Typing.Column(
                name="user_id",
                data_type="INTEGER",
                column_constraints=[c.PrimaryKeyConstraint("User", "user_id")],
            ),
            SQLite3Typing.Column(
                name="user_name",
                data_type="TEXT",
                column_constraints=[c.NotNullConstraint("User", "user_name")],
            ),
        ]

        table = SQLite3Typing.Table(name="User", columns=columns)

    class NewNamePost(Table):  # type: ignore
        columns = [
            SQLite3Typing.Column(
                name="post_id",
                data_type="INTEGER",
                column_constraints=[c.PrimaryKeyConstraint("NewNamePost", "post_id")],
            ),
            SQLite3Typing.Column(
                name="post_name",
                data_type="TEXT",
                column_constraints=[c.NotNullConstraint("NewNamePost", "post_name")],
            ),
        ]

        table = SQLite3Typing.Table(name="NewNamePost", columns=columns)

    SQLite3Config(tmp_database)

    m = MigrationElement()
    m.migrate(preview=True)

    assert type(MigrationElement.operations[0]) == RenameTableOperation
    assert len(MigrationElement.operations) == 1


@pytest.mark.parametrize(
    "tmp_database",
    [[SQLite3TablesData.sql_table_user, SQLite3TablesData.sql_table_post]],
    indirect=True,
)
def test_no_operation_needed(
    tmp_database: str, mock_execute_operations: pytest.MonkeyPatch
):
    # TODO: refactor to avoid duplicating the setup multiple times
    class Table(MigrationElement):

        def __init__(self) -> None:
            self.adapter = SQLite3Adapter()

            models = Table.__subclasses__()
            if self.__class__ == Table:
                for cls in models:
                    self.adapter.create_table(cls.__dict__["table"])

    class User(Table):  # type: ignore
        columns = [
            SQLite3Typing.Column(
                name="user_id",
                data_type="INTEGER",
                column_constraints=[c.PrimaryKeyConstraint("User", "user_id")],
            ),
            SQLite3Typing.Column(
                name="user_name",
                data_type="TEXT",
                column_constraints=[c.NotNullConstraint("User", "user_name")],
            ),
        ]

        table = SQLite3Typing.Table(name="User", columns=columns)

    class Post(Table):  # type: ignore
        columns = [
            SQLite3Typing.Column(
                name="post_id",
                data_type="INTEGER",
                column_constraints=[c.PrimaryKeyConstraint("Post", "post_id")],
            ),
            SQLite3Typing.Column(
                name="post_name",
                data_type="TEXT",
                column_constraints=[c.NotNullConstraint("Post", "post_name")],
            ),
        ]

        table = SQLite3Typing.Table(name="Post", columns=columns)

    SQLite3Config(tmp_database)

    m = MigrationElement()
    m.migrate(preview=True)

    assert len(MigrationElement.operations) == 0


@pytest.mark.parametrize(
    "tmp_database",
    [[SQLite3TablesData.sql_table_user, SQLite3TablesData.sql_table_post]],
    indirect=True,
)
def test_create_table_operation_needed(
    tmp_database: str, mock_execute_operations: pytest.MonkeyPatch
):

    class Table(MigrationElement):

        def __init__(self) -> None:
            self.adapter = SQLite3Adapter()

            models = Table.__subclasses__()
            if self.__class__ == Table:
                for cls in models:
                    self.adapter.create_table(cls.__dict__["table"])

    class User(Table):  # type: ignore
        columns = [
            SQLite3Typing.Column(
                name="user_id",
                data_type="INTEGER",
                column_constraints=[c.PrimaryKeyConstraint("User", "user_id")],
            ),
            SQLite3Typing.Column(
                name="user_name",
                data_type="TEXT",
                column_constraints=[c.NotNullConstraint("User", "user_name")],
            ),
        ]

        table = SQLite3Typing.Table(name="User", columns=columns)

    class Post(Table):  # type: ignore
        columns = [
            SQLite3Typing.Column(
                name="post_id",
                data_type="INTEGER",
                column_constraints=[c.PrimaryKeyConstraint("Post", "post_id")],
            ),
            SQLite3Typing.Column(
                name="post_name",
                data_type="TEXT",
                column_constraints=[c.NotNullConstraint("Post", "post_name")],
            ),
        ]

        table = SQLite3Typing.Table(name="Post", columns=columns)

    class Comment(Table):  # type: ignore

        columns = [
            SQLite3Typing.Column(
                name="comment_id",
                data_type="INTEGER",
                column_constraints=[c.PrimaryKeyConstraint("Comment", "comment_id")],
            ),
            SQLite3Typing.Column(
                name="comment_name",
                data_type="TEXT",
                column_constraints=[c.NotNullConstraint("Comment", "comment_name")],
            ),
        ]

        table = SQLite3Typing.Table(name="Comment", columns=columns)

    SQLite3Config(tmp_database)

    m = MigrationElement()
    m.migrate(preview=True)

    assert type(MigrationElement.operations[0]) == CreateTableOperation
    assert len(MigrationElement.operations) == 1


@pytest.mark.parametrize(
    "tmp_database",
    [
        [
            SQLite3TablesData.sql_table_user,
            SQLite3TablesData.sql_table_post,
            SQLite3TablesData.sql_table_comment,
        ]
    ],
    indirect=True,
)
def test_delete_table_operation_needed(
    tmp_database: str, mock_execute_operations: pytest.MonkeyPatch
):

    class Table(MigrationElement):

        def __init__(self) -> None:
            self.adapter = SQLite3Adapter()

            models = Table.__subclasses__()
            if self.__class__ == Table:
                for cls in models:
                    self.adapter.create_table(cls.__dict__["table"])

    class User(Table):  # type: ignore
        columns = [
            SQLite3Typing.Column(
                name="user_id",
                data_type="INTEGER",
                column_constraints=[c.PrimaryKeyConstraint("User", "user_id")],
            ),
            SQLite3Typing.Column(
                name="user_name",
                data_type="TEXT",
                column_constraints=[c.NotNullConstraint("User", "user_name")],
            ),
        ]

        table = SQLite3Typing.Table(name="User", columns=columns)

    class Post(Table):  # type: ignore
        columns = [
            SQLite3Typing.Column(
                name="post_id",
                data_type="INTEGER",
                column_constraints=[c.PrimaryKeyConstraint("Post", "post_id")],
            ),
            SQLite3Typing.Column(
                name="post_name",
                data_type="TEXT",
                column_constraints=[c.NotNullConstraint("Post", "post_name")],
            ),
        ]

        table = SQLite3Typing.Table(name="Post", columns=columns)

    SQLite3Config(tmp_database)

    m = MigrationElement()
    m.migrate(preview=True)

    assert len(MigrationElement.operations) == 1
    assert type(MigrationElement.operations[0]) == DeleteTableOperation


@pytest.mark.parametrize(
    "tmp_database",
    [[SQLite3TablesData.sql_table_user, SQLite3TablesData.sql_table_post]],
    indirect=True,
)
def test_create_column_operation_needed(
    tmp_database: str, mock_execute_operations: pytest.MonkeyPatch
):

    class Table(MigrationElement):

        def __init__(self) -> None:
            self.adapter = SQLite3Adapter()

            models = Table.__subclasses__()
            if self.__class__ == Table:
                for cls in models:
                    self.adapter.create_table(cls.__dict__["table"])

    class User(Table):  # type: ignore
        columns = [
            SQLite3Typing.Column(
                name="user_id",
                data_type="INTEGER",
                column_constraints=[c.PrimaryKeyConstraint("User", "user_id")],
            ),
            SQLite3Typing.Column(
                name="user_name",
                data_type="TEXT",
                column_constraints=[c.NotNullConstraint("User", "user_name")],
            ),
        ]

        table = SQLite3Typing.Table(name="User", columns=columns)

    class Post(Table):  # type: ignore
        columns = [
            SQLite3Typing.Column(
                name="post_id",
                data_type="INTEGER",
                column_constraints=[c.PrimaryKeyConstraint("Post", "post_id")],
            ),
            SQLite3Typing.Column(
                name="post_name",
                data_type="TEXT",
                column_constraints=[c.NotNullConstraint("Post", "post_name")],
            ),
            SQLite3Typing.Column(
                name="post_content",
                data_type="TEXT",
                column_constraints=[c.NotNullConstraint("Post", "post_content")],
            ),
        ]

        table = SQLite3Typing.Table(name="Post", columns=columns)

    SQLite3Config(tmp_database)

    m = MigrationElement()
    m.migrate(preview=True)

    assert type(MigrationElement.operations[0]) == AddColumnOperation
    assert len(MigrationElement.operations) == 1


@pytest.mark.parametrize(
    "tmp_database",
    [[SQLite3TablesData.sql_table_user, SQLite3TablesData.sql_table_post]],
    indirect=True,
)
def test_delete_column_operation_needed(
    tmp_database: str, mock_execute_operations: pytest.MonkeyPatch
):

    class Table(MigrationElement):

        def __init__(self) -> None:
            self.adapter = SQLite3Adapter()

            models = Table.__subclasses__()
            if self.__class__ == Table:
                for cls in models:
                    self.adapter.create_table(cls.__dict__["table"])

    class User(Table):  # type: ignore
        columns = [
            SQLite3Typing.Column(
                name="user_id",
                data_type="INTEGER",
                column_constraints=[c.PrimaryKeyConstraint("User", "user_id")],
            ),
            SQLite3Typing.Column(
                name="user_name",
                data_type="TEXT",
                column_constraints=[c.NotNullConstraint("User", "user_name")],
            ),
        ]

        table = SQLite3Typing.Table(name="User", columns=columns)

    class Post(Table):  # type: ignore
        columns = [
            SQLite3Typing.Column(
                name="post_id",
                data_type="INTEGER",
                column_constraints=[c.PrimaryKeyConstraint("Post", "post_id")],
            ),
        ]

        table = SQLite3Typing.Table(name="Post", columns=columns)

    SQLite3Config(tmp_database)

    m = MigrationElement()
    m.migrate(preview=True)

    assert type(MigrationElement.operations[0]) == DeleteColumnOperation
    assert len(MigrationElement.operations) == 1


@pytest.mark.parametrize(
    "tmp_database",
    [
        [
            SQLite3TablesData.sql_table_user,
            SQLite3TablesData.sql_table_post,
            SQLite3TablesData.sql_table_comment,
        ]
    ],
    indirect=True,
)
def test_add_column_and_delete_tabe_operation_needed(
    tmp_database: str, mock_execute_operations: pytest.MonkeyPatch
):
    class Table(MigrationElement):

        def __init__(self) -> None:
            self.adapter = SQLite3Adapter()

            models = Table.__subclasses__()
            if self.__class__ == Table:
                for cls in models:
                    self.adapter.create_table(cls.__dict__["table"])

    class User(Table):  # type: ignore
        columns = [
            SQLite3Typing.Column(
                name="user_id",
                data_type="INTEGER",
                column_constraints=[c.PrimaryKeyConstraint("User", "user_id")],
            ),
            SQLite3Typing.Column(
                name="user_name",
                data_type="TEXT",
                column_constraints=[c.NotNullConstraint("User", "user_name")],
            ),
        ]

        table = SQLite3Typing.Table(name="User", columns=columns)

    class Post(Table):  # type: ignore
        columns = [
            SQLite3Typing.Column(
                name="post_id",
                data_type="INTEGER",
                column_constraints=[c.PrimaryKeyConstraint("Post", "post_id")],
            ),
            SQLite3Typing.Column(
                name="post_name",
                data_type="TEXT",
                column_constraints=[c.NotNullConstraint("Post", "post_name")],
            ),
            SQLite3Typing.Column(
                name="post_content",
                data_type="TEXT",
                column_constraints=[c.NotNullConstraint("Post", "post_content")],
            ),
        ]

        table = SQLite3Typing.Table(name="Post", columns=columns)

    SQLite3Config(tmp_database)

    m = MigrationElement()
    m.migrate(preview=True)

    assert type(MigrationElement.operations[1]) == AddColumnOperation
    assert type(MigrationElement.operations[0]) == DeleteTableOperation

    assert len(MigrationElement.operations) == 2



from skibidi_orm.migration_engine.adapters.providers import DatabaseProvider
from skibidi_orm.migration_engine.converters.sqlite3.all import SQLite3Converter
from skibidi_orm.migration_engine.db_config.base_config import BaseDbConfig
from skibidi_orm.migration_engine.revisions.revision import Revision
from skibidi_orm.migration_engine.sql_executor.sqlite3_executor import SQLite3Executor
import sqlite3

from skibidi_orm.migration_engine.state_manager.state_manager import StateManager


class RevisionManager:
    """Class responsible for moving between revisions"""

    type RevisionRepository = dict[int, Revision]

    revisions: RevisionRepository

    def __init__(self):
        config_instance = BaseDbConfig.get_instance()
        database_provider = config_instance.database_provider

        if database_provider == DatabaseProvider.SQLITE3:
            self.converter = SQLite3Converter
            self.executor = SQLite3Executor
        else:
            raise NotImplementedError()

        if not self.find_revision_table():
            self.create_revision_table()

        self.revisions = self.get_all_revisions()

    def find_revision_table(self) -> bool:
        """Used to find the revision table with the name
        specified by the user. Returns True if found,
        False otherwise"""
        try:
            revision_data_query = (
                self.converter.get_query_converter().get_revision_data_query()
            )
            self.executor.execute_sql(revision_data_query)
            return True
        except (
            sqlite3.OperationalError  # more exceptions can be added here for each provider
        ):
            return False

    def create_revision_table(self) -> None:
        """If the revision table is not found when migrating,
        this is called to create it"""
        self.executor.execute_sql(
            self.converter.get_table_operation_converter().get_revision_table_creation_query()
        )

    def save_revision(self, revision: Revision) -> None:
        """Saves a given revision to the revision database"""
        self.executor.save_revision(revision)

    def get_all_revisions(self) -> RevisionRepository:
        """Update and return all the revisions in the database"""
        self.revisions = {
            id: revision for id, revision in self.executor.get_all_revisions()
        }
        return self.revisions

    def go_to_revision(self, revision: Revision) -> None:
        """Checkouts the user to a given revision.
        THIS SHOULD BE ONLY PERFORMED ON AN EMPTY TABLE!"""
        self.clear_database()
        sql_to_execute = self.get_revision_SQL(revision)
        self.executor.execute_sql(sql_to_execute)

    def get_revision_by_id(self, revision_id: int) -> Revision:
        """Returns a revision object by its id. If not found,
        throws an exception"""
        return self.revisions[revision_id]

    def get_revision_SQL(self, revision: Revision) -> str:
        """Returns the SQL string, the execution of which
        will result in checkouting the revision"""
        tables = revision.tables
        # initialize the statemanager with an empty list, so that
        # everything is considered as a new table
        state_manager = StateManager([], tables)
        operations = state_manager.operations
        return "\n".join(
            (
                self.converter.convert_operation_to_SQL(operation)
                for operation in operations
            )
        )

    def clear_database(self) -> None:
        """Clears the database of all tables except the revision table.
        Used before migrating to a revision"""
        query = self.converter.get_query_converter().get_table_clearing_query()
        self.executor.execute_sql(query)



from abc import ABC, abstractmethod
from typing import Any

from skibidi_orm.migration_engine.adapters.base_adapter import (
    BaseColumn,
    BaseTable,
)


class BaseDbInspector(ABC):
    @abstractmethod
    def get_tables(self) -> list[BaseTable[BaseColumn[Any]]]:
        """
        Get all tables from the database.
        """
        pass

    @abstractmethod
    def get_tables_names(self) -> list[str]:
        """
        Get all tables names from the database.
        """
        pass

    @abstractmethod
    def get_table_columns(self, table_name: str) -> list[BaseColumn[Any]]:
        """
        Get all columns from a table.
        """
        pass



from skibidi_orm.exceptions.constraints import UnsupportedConstraintError
from skibidi_orm.migration_engine.converters.base.interfaces import (
    ConstraintSQLConverter,
)
from skibidi_orm.migration_engine.adapters.database_objects.constraints import (
    CheckConstraint,
    Constraint,
    ConstraintType,
    ForeignKeyConstraint,
)
from typing import cast


class PostgresConstraintConverter(ConstraintSQLConverter):
    """Class responsible for converting constraint objects to raw Postgres SQL strings"""

    @staticmethod
    def convert_constraint_change_operation_to_SQL(_: Constraint) -> str:
        """Convert a given constraint change operation to a Postgres SQL string"""
        raise NotImplementedError(
            "Constraint change operations are not supported in Postgres yet."
        )

    @staticmethod
    def convert_constraint_to_SQL(constraint: Constraint) -> str:
        """Convert a given constraint to a Postgres SQL string"""
        if constraint.constraint_type == ConstraintType.PRIMARY_KEY:
            return "PRIMARY KEY"
        elif constraint.constraint_type == ConstraintType.UNIQUE:
            return "UNIQUE"
        elif constraint.constraint_type == ConstraintType.FOREIGN_KEY:
            constraint = cast(ForeignKeyConstraint, constraint)
            return (
                f"FOREIGN KEY ({', '.join(constraint.column_mapping.keys())}) REFERENCES"
                f" {constraint.referenced_table} ({', '.join(constraint.column_mapping.values())})"
            )
        elif constraint.constraint_type == ConstraintType.CHECK:
            return f"CHECK ({cast(CheckConstraint, constraint).column_name} {cast(CheckConstraint, constraint).condition})"
        elif constraint.constraint_type == ConstraintType.NOT_NULL:
            return "NOT NULL"
        else:
            raise UnsupportedConstraintError(
                f"Constraints of type {constraint.constraint_type} are not supported by Postgres"
            )



from skibidi_orm.migration_engine.adapters.database_objects.constraints import (
    ForeignKeyConstraint,
)
from skibidi_orm.migration_engine.adapters.sqlite3_typing import (
    SQLite3Typing,
)
from skibidi_orm.migration_engine.adapters.providers import DatabaseProvider
from skibidi_orm.migration_engine.db_config.sqlite3_config import SQLite3Config
from skibidi_orm.migration_engine.db_inspectors.sqlite3_inspector import (
    SQLite3Inspector,
)
from skibidi_orm.migration_engine.revisions.manager import RevisionManager
from skibidi_orm.migration_engine.revisions.revision import Revision
import pytest

sql_schema_with_fks = [
    # todo: remove after postgres is merged
    """
    CREATE TABLE users (
        user_id INTEGER PRIMARY KEY,
        username TEXT NOT NULL UNIQUE,
        email TEXT NOT NULL UNIQUE,
        password_hash TEXT NOT NULL,
        registration_date TIMESTAMP DEFAULT CURRENT_TIMESTAMP
    );
""",
    """
    CREATE TABLE posts (
        post_id INTEGER PRIMARY KEY,
        user_id INTEGER NOT NULL,
        title TEXT NOT NULL,
        content TEXT NOT NULL,
        post_date TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
        FOREIGN KEY (user_id) REFERENCES users(user_id)
    );
""",
    """
    CREATE TABLE comments (
        comment_id INTEGER PRIMARY KEY,
        username TEXT NOT NULL,
        user_idd INTEGER NOT NULL,
        post_id INTEGER NOT NULL,
        comment_text TEXT NOT NULL,
        comment_date TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
        FOREIGN KEY (user_idd, username) REFERENCES users(user_id, username),
        FOREIGN KEY (post_id) REFERENCES posts(post_id)
    );
""",
]

simple_testing_schema = [
    "CREATE TABLE test_table (id INTEGER, name TEXT);",
    "CREATE TABLE test_table_2 (idd INTEGER, count_ INTEGER, FOREIGN KEY (idd) REFERENCES test_table (id));",
    "INSERT INTO test_table (id, name) VALUES (1, 'test');",
]

table_1 = SQLite3Typing.Table(
    "test_table",
    [SQLite3Typing.Column("id", "INTEGER"), SQLite3Typing.Column("name", "TEXT")],
)

table_2 = SQLite3Typing.Table(
    "test_table_2",
    [
        SQLite3Typing.Column("idd", "INTEGER"),
        SQLite3Typing.Column("count_", "INTEGER"),
    ],
    {ForeignKeyConstraint("test_table_2", "test_table", {"idd": "id"})},
)

revision_1 = Revision(
    "test description",
    "test schema repr",
    DatabaseProvider.SQLITE3,
    [table_1, table_2],
)

revision_2 = Revision(
    "test description_2",
    "test schema repr_2",
    DatabaseProvider.SQLITE3,
    [SQLite3Typing.Table("test_table", [])],
)


def test_create_and_find_revision_table_sqlite(
    make_database: str, monkeypatch: pytest.MonkeyPatch
):
    """Tests whether the revision table is found if it does not exist"""
    monkeypatch.setattr(RevisionManager, "create_revision_table", lambda _: None)  # type: ignore
    monkeypatch.setattr(RevisionManager, "get_all_revisions", lambda _: [])  # type: ignore

    SQLite3Config(make_database)
    manager = RevisionManager()
    assert not manager.find_revision_table()


def test_save_revision_sqlite(make_database: str):
    """Tests whether a revision can be saved without causing
    any database errors"""
    SQLite3Config(make_database)
    manager = RevisionManager()

    if not manager.find_revision_table():
        manager.create_revision_table()
    manager.save_revision(revision_1)


def test_save_and_get_revision_sqlite(make_database: str):
    """Tests whether a revision can be saved and then retrieved
    from the database"""
    SQLite3Config(make_database)
    manager = RevisionManager()

    if not manager.find_revision_table():
        manager.create_revision_table()

    manager.save_revision(revision_1)
    manager.save_revision(revision_2)

    revisions = manager.get_all_revisions()

    assert len(revisions) == 2
    assert revisions[1] == revision_1
    assert revisions[2] == revision_2


@pytest.mark.parametrize("make_database", [simple_testing_schema], indirect=True)
def test_clear_database_sqlite(make_database: str):
    """Clearing the database removes everything but the revisions table and its contents"""
    SQLite3Config(make_database)
    manager = RevisionManager()

    manager.save_revision(revision_1)
    manager.clear_database()

    revisions = manager.get_all_revisions()
    assert len(revisions) == 1
    revision_ = revisions.values().__iter__().__next__()
    assert revision_ == revision_1


@pytest.mark.parametrize("make_database", [simple_testing_schema], indirect=True)
def test_get_revision_SQL_sqlite(make_database: str):
    SQLite3Config(make_database)
    manager = RevisionManager()

    revision_sql = manager.get_revision_SQL(revision_1)
    assert revision_sql == (
        "CREATE TABLE test_table (id INTEGER, name TEXT);\n"
        "CREATE TABLE test_table_2 (idd INTEGER, count_ INTEGER, FOREIGN KEY (idd) "
        "REFERENCES test_table (id));"
    )


def test_go_to_revision_from_empty_db_sqlite(make_database: str):
    """Tests whether the go_to_revision method works when the database is empty"""
    SQLite3Config(make_database)

    manager = RevisionManager()
    manager.save_revision(revision_1)
    revision_id = manager.get_all_revisions().keys().__iter__().__next__()

    assert manager.get_revision_by_id(revision_id) == revision_1
    manager.go_to_revision(revision_1)

    all_tables = SQLite3Inspector().get_tables()

    assert len(all_tables) == 2
    assert table_1 in all_tables
    assert table_2 in all_tables


@pytest.mark.parametrize("make_database", [sql_schema_with_fks], indirect=True)
def test_go_to_revision_from_non_empty_db_sqlite(make_database: str):
    SQLite3Config(make_database)
    inspector = SQLite3Inspector()
    assert len(inspector.get_tables()) == 3

    manager = RevisionManager()
    manager.go_to_revision(revision_1)
    new_tables = inspector.get_tables()
    assert table_1 in new_tables
    assert table_2 in new_tables
    assert len(new_tables) == 2



import pytest

from skibidi_orm.migration_engine.adapters.postgres_typing import PostgresTyping
from skibidi_orm.migration_engine.converters.postgres.tables import (
    PostgresTableOperationConverter,
)
from skibidi_orm.migration_engine.adapters.database_objects.constraints import (
    CheckConstraint,
    ForeignKeyConstraint,
    PrimaryKeyConstraint,
    UniqueConstraint,
    NotNullConstraint,
)
from skibidi_orm.migration_engine.operations.table_operations import (
    CreateTableOperation,
    DeleteTableOperation,
    RenameTableOperation,
)


simple_table_no_constraints = PostgresTyping.Table(
    "users", columns=[PostgresTyping.Column("name", "TEXT")]
)


complex_table_no_constraints = PostgresTyping.Table(
    "users",
    columns=[
        PostgresTyping.Column("name", "TEXT"),
        PostgresTyping.Column("age", "INTEGER"),
        PostgresTyping.Column("email", "TEXT"),
        PostgresTyping.Column("active", "BYTEA"),
    ],
)

complex_table_with_constraints = PostgresTyping.Table(
    "admin_users",
    columns=[
        PostgresTyping.Column(
            "user_id",
            "INTEGER",
            column_constraints=[
                PrimaryKeyConstraint("users", "user_id"),
                NotNullConstraint("users", "user_id"),
            ],
        ),
        PostgresTyping.Column("name", "TEXT"),
        PostgresTyping.Column(
            "email", "TEXT", column_constraints=[UniqueConstraint("users", "email")]
        ),
        PostgresTyping.Column("active", "BYTEA", column_constraints=list()),
        PostgresTyping.Column(
            "age",
            "INTEGER",
            column_constraints=[CheckConstraint("users", "age", "> 18")],
        ),
    ],
    foreign_keys={
        ForeignKeyConstraint(
            "admin_users", "users", {"active": "active", "name": "name"}
        )
    },
)


@pytest.mark.parametrize(
    "operation, expected",
    [
        (
            CreateTableOperation(simple_table_no_constraints),
            "CREATE TABLE users (name TEXT);",
        ),
        (
            CreateTableOperation(complex_table_no_constraints),
            "CREATE TABLE users (name TEXT, age INTEGER, email TEXT, active BYTEA);",
        ),
        (
            CreateTableOperation(complex_table_with_constraints),
            "CREATE TABLE admin_users (user_id INTEGER PRIMARY KEY NOT NULL, name TEXT, "
            "email TEXT UNIQUE, active BYTEA, age INTEGER, CHECK (age > 18), FOREIGN KEY (active, name) REFERENCES users ("
            "active, name));",
        ),
        (
            DeleteTableOperation(simple_table_no_constraints),
            "DROP TABLE users;",
        ),
        (
            RenameTableOperation(simple_table_no_constraints, "users2"),
            "DROP TABLE users; CREATE TABLE users2 (name TEXT);",
        ),
    ],
)
def test_convert_table_operation_to_SQL(operation, expected):  # type: ignore
    """Test the conversion of a CreateTableOperation to SQL."""
    assert (
        PostgresTableOperationConverter.convert_table_operation_to_SQL(operation)  # type: ignore
        == expected
    )



from abc import ABC, abstractmethod
from typing import Any

from skibidi_orm.migration_engine.operations.table_operations import TableOperation
from skibidi_orm.migration_engine.operations.column_operations import ColumnOperation
from skibidi_orm.migration_engine.revisions.revision import Revision


class BaseSQLExecutor(ABC):

    @staticmethod
    @abstractmethod
    def execute_sql(sql: str) -> None:
        """Execute the given SQL query in the database"""

    @staticmethod
    @abstractmethod
    def execute_sql_query(sql: str) -> list[Any]:
        """Execute the given SQL query in the database and
        return its result"""

    @staticmethod
    @abstractmethod
    def execute_operations(operations: list[TableOperation | ColumnOperation]) -> None:
        """Execute the given operations in the database"""

    @staticmethod
    @abstractmethod
    def save_revision(revision: Revision) -> None:
        """Save the given revision in the database"""

    @staticmethod
    @abstractmethod
    def get_all_revisions() -> list[tuple[int, Revision]]:
        """Get all the revisions from the database
        returns a list of tuples (id, revision)"""



from __future__ import annotations
from abc import abstractmethod, ABC
from skibidi_orm.migration_engine.operations.column_operations import ColumnOperation
from skibidi_orm.migration_engine.operations.table_operations import (
    TableOperation,
)
from skibidi_orm.migration_engine.adapters.database_objects.constraints import (
    Constraint,
)


class SQLConverter(ABC):
    """Class responsible for converting operation and constraint objects
    to raw SQL strings in a specific dialect. Each of the supported database
    providers should inherit a converter from this base class and implement its
    methods"""

    @staticmethod
    @abstractmethod
    def get_column_operation_converter() -> type[ColumnOperationSQLConverter]:
        """Return the corresponding column operation converter class"""

    @staticmethod
    @abstractmethod
    def get_table_operation_converter() -> type[TableOperationSQLConverter]:
        """Return the corresponsing table operation converter class"""

    @classmethod
    def convert_operation_to_SQL(
        cls, operation: ColumnOperation | TableOperation
    ) -> str:
        """Convert a given operation object to raw SQL in a specific dialect"""
        if isinstance(operation, ColumnOperation):
            return cls.get_column_operation_converter().convert_column_operation_to_SQL(
                operation
            )
        return cls.get_table_operation_converter().convert_table_operation_to_SQL(
            operation
        )

    @staticmethod
    @abstractmethod
    def get_query_converter() -> type[SQLQueryConverter]:
        """Return the corresponding query converter class"""

    @classmethod
    def get_revision_table_creation_query(cls) -> str:
        """Return the SQL string which creates a special internal table
        used to hold revision data"""
        return cls.get_table_operation_converter().get_revision_table_creation_query()

    @classmethod
    @abstractmethod
    def get_table_clearing_query(cls) -> str:
        """Return the SQL string which clears the table of all data
        except the revision table"""
        return cls.get_query_converter().get_table_clearing_query()

    @classmethod
    @abstractmethod
    def get_revision_insertion_query(cls) -> str:
        """Return the SQL string which inserts a revision into the revision table.
        The revision needs to be provided as a second argument to cur.execute()"""

    @classmethod
    def get_revision_data_query(cls) -> str:
        """Return the SQL string which selects all of the data from the revision table"""
        return cls.get_query_converter().get_revision_data_query()


class ColumnOperationSQLConverter(ABC):
    """Class responsible for converting column operation objects to raw SQL strings
    in a specific dialect. Each of the supported database providers should inherit
    a converter from this base class and implement its methods"""

    @staticmethod
    @abstractmethod
    def convert_column_operation_to_SQL(operation: ColumnOperation) -> str:
        """Convert a given column operation object to raw SQL in a specific dialect"""


class TableOperationSQLConverter(ABC):
    """Class responsible for converting table operation objects to raw SQL strings
    in a specific dialect. Each of the supported database providers should inherit
    a converter from this base class and implement its methods"""

    @staticmethod
    @abstractmethod
    def convert_table_operation_to_SQL(operation: TableOperation) -> str:
        """Convert a given table operation object to raw SQL in a specific dialect"""

    @staticmethod
    @abstractmethod
    def get_revision_table_creation_query() -> str:
        """Return the SQL string which creates a special internal table
        used to hold revision data"""


class ConstraintSQLConverter(ABC):
    """Class responsible for converting constraint objects to raw SQL strings
    in a specific dialect. Each of the supported database providers should inherit
    a converter from this base class and implement its methods"""

    @staticmethod
    @abstractmethod
    def convert_constraint_to_SQL(constraint: Constraint) -> str:
        """Convert a given constraint object to raw SQL in a specific dialect"""


class SQLQueryConverter(ABC):
    """Class responsible for converting SELECT queries to raw SQL strings.
    Used mainly to retrieve data from the revision table."""

    @staticmethod
    @abstractmethod
    def get_revision_data_query() -> str:
        """Return the SQL string which selects all of the data from the revision table"""

    @staticmethod
    @abstractmethod
    def get_table_clearing_query() -> str:
        """Return the SQL string which clears the table of all data
        except the revision table"""



import pytest

from skibidi_orm.migration_engine.data_mutator.postgres_data_mutator import (
    PostgresDataMutator,
)
from ..conftest import postgres_db_fixture
from skibidi_orm.migration_engine.db_config.postgres_config import PostgresConfig
from skibidi_orm.migration_engine.db_inspectors.postgres_inspector import (
    PostgresInspector,
)


@pytest.mark.parametrize(
    "create_query, query, expected",
    [
        (
            [
                "CREATE TABLE test (id SERIAL PRIMARY KEY, name TEXT);",
                "INSERT INTO test (name) VALUES ('test');",
            ],
            "SELECT * FROM test;",
            [
                (1, "test"),
            ],
        ),
        (
            ["CREATE TABLE test (id SERIAL PRIMARY KEY, name TEXT);"],
            "INSERT INTO test (name) VALUES ('test');",
            [],
        ),
        (
            ["CREATE TABLE test (id SERIAL PRIMARY KEY, name TEXT);"],
            "SELECT * FROM test;",
            [],
        ),
    ],
)
def test_raw_query(create_query, query, expected):  # type: ignore
    @postgres_db_fixture(
        db_name="postgres",
        db_user="admin",
        db_password="admin",
        db_host="0.0.0.0",
        db_port=5432,
        queries=create_query,  # type: ignore
    )
    def test_fn(config: PostgresConfig, inspector: PostgresInspector):
        mutator = PostgresDataMutator()
        result = mutator.raw_query(query)  # type: ignore
        assert result == expected

    test_fn()


@pytest.mark.parametrize(
    "create_query, table_name, limit, offset, expected",
    [
        (
            ["CREATE TABLE test (id SERIAL PRIMARY KEY, name TEXT);"],
            "test",
            100,
            0,
            [],
        ),
        (
            [
                "CREATE TABLE test (id SERIAL PRIMARY KEY, name TEXT);",
                "INSERT INTO test (name) VALUES ('test');",
            ],
            "test",
            100,
            0,
            [(1, "test")],
        ),
        (
            [
                "CREATE TABLE test (id SERIAL PRIMARY KEY, name TEXT);",
                "INSERT INTO test (name) VALUES ('test');",
            ],
            "test",
            1,
            1,
            [],
        ),
        (
            [
                "CREATE TABLE test (id SERIAL PRIMARY KEY, name TEXT);",
                "INSERT INTO test (name) VALUES ('test');",
            ],
            "test",
            0,
            0,
            [],
        ),
    ],
)
def test_get_rows(create_query, table_name, limit, offset, expected):  # type: ignore
    @postgres_db_fixture(
        db_name="postgres",
        db_user="admin",
        db_password="admin",
        db_host="0.0.0.0",
        db_port=5432,
        queries=create_query,  # type: ignore
    )
    def test_fn(config: PostgresConfig, inspector: PostgresInspector):
        mutator = PostgresDataMutator()
        results = mutator.get_rows(table_name=table_name, limit=limit, offset=offset)  # type: ignore
        assert sorted(results) == sorted(expected)  # type: ignore

    test_fn()



import pytest
from skibidi_orm.migration_engine.db_config.postgres_config import PostgresConfig
from skibidi_orm.migration_engine.db_config.sqlite3_config import SQLite3Config
import os


def test_should_raise_error_on_invalid_insantiation():
    """
    We cannot instantiate more than one config
    """
    with pytest.raises(RuntimeError) as exc_info:
        SQLite3Config(db_path="first_path")
        SQLite3Config(db_path="next_path")
    assert str(exc_info.value) == "Only one instance of this class is allowed"


def test_should_raise_error_when_instantiating_other_config():
    """
    Should raise error when trying to instantiate other config after one has been instantiated
    """
    SQLite3Config(db_path="first_path")
    with pytest.raises(RuntimeError) as exc_info:
        PostgresConfig(
            db_name="db_name",
            db_user="db_user",
            db_password="db_password",
            db_host="db_host",
            db_port=5432,
        )
    assert str(exc_info.value) == "Only one instance of this class is allowed"


def test_accessing_through_instance_method_should_throw_error_when_instance_does_not_exist():
    """
    Should raise error when trying to access instance method when instance does not exist
    """
    with pytest.raises(ReferenceError) as exc_info:
        SQLite3Config.get_instance()
    assert str(exc_info.value) == "Instance does not exist"


def test_check_argument_saving():
    """
    Should raise error when arguments are not saved
    """
    config = SQLite3Config(db_path="first_path")
    assert config.db_path == os.path.join(os.getcwd(), "first_path")



# This file can by run by executing `skibidi-orm` in the terminal, after the package is installed.
# `skibidi-orm migrate [--message <STRING>]`
# `skibidi-orm list-migrations`
# `skibidi-orm go <MIGRATION_ID>`


from pathlib import Path
import typer
from typing import Union
import os

# import shutil
import sys
from skibidi_orm.cli.migration_file_creator import create_migration_file
from skibidi_orm.cli.utils import find_schema_file, load_schema_from_path
from skibidi_orm.exceptions.cli_exceptions import MultipleSchemaFilesError
from skibidi_orm.migration_engine.adapters.database_objects.migration_element import (
    MigrationElement,
)
from skibidi_orm.migration_engine.studio.server import run_server
from colorama import Fore

sys.path.insert(0, os.getcwd())


app = typer.Typer(
    help="CLI tool for managing schema creations and migrations in Skibidi ORM."
)
state = {"schema_path": Path()}


@app.callback()
def schema_callback(
    schema_path: Union[None, Path] = typer.Option(
        None,
        "--schema-file",
        "-s",
        help="Specify an external schema file path",
        show_default=False,
    )
):
    """
    Specify a path to the schema file to use.
    """

    if schema_path is not None:
        state["schema_path"] = schema_path

        try:
            load_schema_from_path(str(schema_path))
            print("\nSchema file successfully loaded.\n")

        except FileNotFoundError:
            print("\nSchema file could not be found. Aborting.\n")
            raise typer.Exit(code=1)

        except ImportError:
            print("\nSchema file could not be loaded. Aborting.\n")
            raise typer.Exit(code=1)

    else:
        try:
            import schema as schema  # type: ignore

            print("\nSchema file successfully loaded.\n")

        except ImportError:
            print("\nSchema file could not be loaded. Aborting.\n")
            raise typer.Exit(code=1)


@app.command()
def migrate(
    message: str = typer.Option(
        None,
        "--message",
        "-m",
        help="Description of migration",
    ),
    direct: bool = typer.Option(
        False,
        "--direct",
        "-d",
        help="Use --direct to run the migration directly. Without the flag, it will create a python migration file.",
    ),
):
    """
    Used to run migration for current schema file. Can accept an optional message as a description of the migration.
    """

    if direct:
        m = MigrationElement()
        m.migrate(preview=False)
        print("\nMigration complete. \n")
    else:
        m = MigrationElement()
        create_migration_file(m)
        print("\nMigration file created. \n")

    pass


@app.command(name="preview-migration")
def preview_migration():
    """
    Preview the migration that will be executed.
    """

    m = MigrationElement()
    m.migrate(preview=True)

    if not m.operations:
        print("No changes to be made.")
    else:
        print("Migration preview:")
        for i, operation in enumerate(m.operations):
            print(f"\t{i+1}) {operation}")
    pass


@app.command(name="list-migrations")
def migrate_list():
    """
    List all made migrations with their descriptions and ID.
    """
    print("Listing all migrations")
    pass


@app.command()
def go(migration_id: Union[str, None] = typer.Argument(None, help="Migration ID")):
    """
    Go back (and forward) to specific migration.
    """

    if migration_id is None:
        migration_id = typer.prompt("Please enter the migration ID:")

    print(f"Going to migration with ID: {migration_id}")
    pass


@app.command()
def studio(
    schema_file: Union[str, None] = typer.Option(
        None, "--schema-file", "-s", help="Schema file path"
    )
):
    """
    Run web UI for CRUD operations on current DB.
    """

    try:
        if schema_file is None:
            schema_file = find_schema_file()
    except MultipleSchemaFilesError:
        print(
            Fore.RED
            + "Multiple schema files found. Please specify the schema file to use: --schema-file <PATH>"
        )
        raise typer.Exit(code=1)

    try:
        run_server(schema_file=schema_file)
    except Exception as e:
        print(Fore.RED + f"Error: {e}")
        raise typer.Exit(code=1)


def main():
    app()


if __name__ == "__main__":
    app()



from skibidi_orm.migration_engine.converters.sqlite3.queries import (
    SQLite3QueryConverter,
)
from skibidi_orm.migration_engine.revisions.constants import get_revision_table_name


def test_revision_data_query_according_to_get_revision_table_name():
    """Tests whether the get_revision_data_query contains the table name from get_revision_table_name"""
    assert (
        SQLite3QueryConverter.get_revision_data_query()
        == f"SELECT rowid, * FROM {get_revision_table_name()};"
    )


def test_table_clearing_query_according_to_get_revision_table_name():
    assert (
        SQLite3QueryConverter.get_table_clearing_query()
        == f"""
            PRAGMA writable_schema = 1;
            DELETE FROM sqlite_master WHERE type IN ('table', 'index', 'trigger') AND name != '{get_revision_table_name()}';
            PRAGMA writable_schema = 0;
            COMMIT;
            VACUUM;
            PRAGMA INTEGRITY_CHECK;
            """
    )
